# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

try:
    import geomodels  # noqa: F401
except ImportError:
    sys.path.insert(0, os.path.abspath(".."))


def get_version(filename="../geomodels/_version.py", strip_extra=False):
    import re

    import packaging.version

    with open(filename) as fd:
        data = fd.read()

    mobj = re.search(
        r"""^__version__\s*=\s*(?P<quote>['"])(?P<version>.*)(?P=quote)""",
        data,
        re.MULTILINE,
    )

    version = packaging.version.parse(mobj.group("version"))

    if strip_extra:
        return version.base_version
    else:
        return str(version)


# -- Project information -----------------------------------------------------

project = "GeoModels"
copyright = "2019-2025, Antonio Valentino"
author = "Antonio Valentino"

# The short X.Y version.
version = get_version(strip_extra=True)

# The full version, including alpha/beta/rc tags
release = get_version(strip_extra=False)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.extlinks",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    # Service links and badges¶
    # 'badge_branch': 'master',  # not supported by Sphinx v1.8.5
    # 'codecov_button': True,
    "github_user": "avalentino",
    "github_repo": "geomodels",
    "github_banner": True,
    "github_button": True,
    "github_type": "watch",  # 'watch', 'fork', 'follow'
    "github_count": True,
    # 'travis_button': True,
    # Non-service sidebar control
    "extra_nav_links": {
        "GeoModels on PyPI": "https://pypi.org/project/geomodels",
        "GeoGraphicLib": "https://github.com/geographiclib/geographiclib",
    },
    # 'show_related': True,
    # 'sidebar_collapse': True,  # not supported by Sphinx v1.8.5
    # Header/footer options
    # 'show_powered_by': True,
    # 'show_relbars': True,  # not supported by Sphinx v1.8.5
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",
        "searchbox.html",
    ],
}


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    "papersize": "a4paper",
    # The font size ('10pt', '11pt' or '12pt').
    "pointsize": "12pt",
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
    # Latex figure (float) alignment
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "GeoModels.tex", "GeoModels Documentation", author, "manual"),
]

# If false, no module index is generated.
latex_domain_indices = False


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    # '/usr/share/doc/python3/html/objects.inv'),
    "numpy": ("https://docs.scipy.org/doc/numpy", None),
    # '/usr/share/doc/python-numpy-doc/html/objects.inv'),
}


# -- Options for extlinks extension ------------------------------------------

# External links configuration
extlinks = {
    "issue": ("https://github.com/avalentino/geomodels/issues/%s", "gh-%s"),
}



# -*- coding: utf-8 -*-
# cython: language_level=3
# distutils: language=c++

from libcpp.string cimport string

from .common cimport real


cdef extern from "GeographicLib/MagneticModel.hpp" namespace "GeographicLib" nogil:
    cdef cppclass CMagneticModel "GeographicLib::MagneticModel":
        CMagneticModel(const string& name, const string& path) except +
        # @TODO: const Geocentric& earth=Geocentric.WGS84()

        void operator()(
            real t, real lat, real lon, real h, real& Bx, real& By, real& Bz
        ) const
        void operator()(
            real t,
            real lat,
            real lon,
            real h,
            real& Bx,
            real& By,
            real& Bz,
            real& Bxt,
            real& Byt,
            real& Bzt,
        ) const

        # MagneticCircle Circle(real t, real lat, real h) const

        # New in GeographicLib v1.52
        # void FieldGeocentric(
        #     real t,
        #     real X,
        #     real Y,
        #     real Z,
        #     real& BX,
        #     real& BY,
        #     real& BZ,
        #     real& BXt,
        #     real& BYt,
        #     real& BZt,
        # ) const

        @staticmethod
        void FieldComponents(
            real Bx, real By, real Bz, real& H, real& F, real& D, real& I
        )

        # @staticmethod
        # void FieldComponents(
        #     real Bx,
        #     real By,
        #     real Bz,
        #     real Bxt,
        #     real Byt,
        #     real Bzt,
        #     real& H,
        #     real& F,
        #     real& D,
        #     real& I,
        #     real& Ht,
        #     real& Ft,
        #     real& Dt,
        #     real& It,
        # )

        const string& Description() const
        const string& DateTime() const
        const string& MagneticFile() const
        const string& MagneticModelName() const
        const string& MagneticModelDirectory() const

        real MinHeight() const
        real MaxHeight() const
        real MinTime() const
        real MaxTime() const
        real EquatorialRadius() const
        real Flattening() const

        int Degree() const
        int Order() const

        @staticmethod
        string DefaultMagneticPath()

        @staticmethod
        string DefaultMagneticName()



import io
import re
import sys
import pathlib
import tempfile
import unittest
import subprocess
from unittest import mock
from contextlib import redirect_stdout

import geomodels.data
from geomodels import cli, get_default_data_path
from geomodels import __package__ as PKG  # noqa: N812
from geomodels import __version__ as VERSION  # noqa: N812


def _make_cmd(*args):
    return [sys.executable, "-m", PKG] + list(args)


class MainTestCase(unittest.TestCase):
    def test_version(self):
        cmd = _make_cmd("--version")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        self.assertIn(VERSION, result.stdout)

    def test_help(self):
        cmd = _make_cmd("--help")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        usage = result.stdout.splitlines()[0]
        self.assertIn("usage:", usage)
        self.assertIn(cli.PROG, usage)

    def test_h(self):
        cmd = _make_cmd("-h")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        usage = result.stdout.splitlines()[0]
        self.assertIn("usage:", usage)
        self.assertIn(cli.PROG, usage)


class InfoSubCommandTestCase(unittest.TestCase):
    VERSION_RE = re.compile(rf"^geomodels version:\s+{VERSION}$", re.MULTILINE)
    DATAPATH_RE = re.compile(
        rf"^data directory:\s+{get_default_data_path()!r}$", re.MULTILINE
    )

    def test_help(self):
        cmd = _make_cmd("info", "--help")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        usage = result.stdout.splitlines()[0]
        self.assertIn("usage:", usage)
        self.assertIn(cli.PROG, usage)
        self.assertIn("info", usage)

    def test_h(self):
        cmd = _make_cmd("info", "-h")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        usage = result.stdout.splitlines()[0]
        self.assertIn("usage:", usage)
        self.assertIn(cli.PROG, usage)
        self.assertIn("info", usage)

    def test_default(self):
        with redirect_stdout(io.StringIO()) as sout:
            ret = cli.main("info")
        self.assertEqual(ret, None)
        output = sout.getvalue()
        self.assertRegex(output, self.VERSION_RE)
        self.assertNotRegex(output, self.DATAPATH_RE)

    def test_data(self):
        with redirect_stdout(io.StringIO()) as sout:
            ret = cli.main("info", "--data")
        self.assertEqual(ret, None)
        output = sout.getvalue()
        self.assertNotRegex(output, self.VERSION_RE)
        self.assertRegex(output, self.DATAPATH_RE)

    def test_all(self):
        with redirect_stdout(io.StringIO()) as sout:
            ret = cli.main("info", "--all")
        self.assertEqual(ret, None)
        output = sout.getvalue()
        self.assertRegex(output, self.VERSION_RE)
        self.assertRegex(output, self.DATAPATH_RE)


class InstallDataSubCommandTestCase(unittest.TestCase):
    MODEL_TYPES = [
        geomodels.data.EGeoidModel,
        geomodels.data.EGravityModel,
        geomodels.data.EMagneticModel,
    ]

    def setUp(self) -> None:
        self.download_patcher = mock.patch(
            "geomodels.data.download", return_value="dummy"
        )
        self.unpack_archive_patcher = mock.patch("shutil.unpack_archive")
        self.download_mock = self.download_patcher.start()
        self.unpack_archive_mock = self.unpack_archive_patcher.start()

    def tearDown(self) -> None:
        self.download_patcher.stop()
        self.unpack_archive_patcher.stop()

    def test_help(self):
        cmd = _make_cmd("install-data", "--help")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        usage = result.stdout.splitlines()[0]
        self.assertIn("usage:", usage)
        self.assertIn(cli.PROG, usage)
        self.assertIn("install-data", usage)

    def test_h(self):
        cmd = _make_cmd("install-data", "-h")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        usage = result.stdout.splitlines()[0]
        self.assertIn("usage:", usage)
        self.assertIn(cli.PROG, usage)
        self.assertIn("install-data", usage)

    def test_install_all_models(self):
        model = geomodels.data.EModelGroup.ALL

        with tempfile.TemporaryDirectory() as datadir:
            ret = cli.main(
                "install-data", "-d", datadir, "--no-progress", model.value
            )

        self.assertEqual(ret, None)

        n_models = sum(len(m) for m in self.MODEL_TYPES)
        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, n_models)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertEqual(self.unpack_archive_mock.call_count, n_models)

    def test_install_minimal_models(self):
        model = geomodels.data.EModelGroup.MINIMAL

        with tempfile.TemporaryDirectory() as datadir:
            ret = cli.main(
                "install-data", "-d", datadir, "--no-progress", model.value
            )

        self.assertEqual(ret, None)

        n_models = 3
        self.download_mock.assert_called()
        self.assertEqual(self.download_mock.call_count, n_models)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertEqual(self.unpack_archive_mock.call_count, n_models)

    def test_install_recommended_models(self):
        model = geomodels.data.EModelGroup.RECOMMENDED

        with tempfile.TemporaryDirectory() as datadir:
            ret = cli.main(
                "install-data", "-d", datadir, "--no-progress", model.value
            )

        self.assertEqual(ret, None)
        self.download_mock.assert_called()
        self.assertGreaterEqual(self.download_mock.call_count, 3)
        self.assertLessEqual(self.download_mock.call_count, 7)
        self.unpack_archive_mock.assert_called_with(
            mock.ANY, extract_dir=datadir
        )
        self.assertGreaterEqual(self.unpack_archive_mock.call_count, 3)
        self.assertLessEqual(self.unpack_archive_mock.call_count, 7)


class ImportIgrfSubCommandTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.wmmdata_mock = mock.Mock()
        self.import_igrf_patcher = mock.patch(
            "geomodels.cli.import_igrf_txt", return_value=self.wmmdata_mock
        )
        self.import_igrf_mock = self.import_igrf_patcher.start()

    def tearDown(self) -> None:
        self.import_igrf_patcher.stop()

    def test_help(self):
        cmd = _make_cmd("import-igrf", "--help")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        usage = result.stdout.splitlines()[0]
        self.assertIn("usage:", usage)
        self.assertIn(cli.PROG, usage)
        self.assertIn("import-igrf", usage)

    def test_h(self):
        cmd = _make_cmd("import-igrf", "-h")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf-8")
        self.assertEqual(result.returncode, 0)
        usage = result.stdout.splitlines()[0]
        self.assertIn("usage:", usage)
        self.assertIn(cli.PROG, usage)
        self.assertIn("import-igrf", usage)

    def test_import_igrf_file(self):
        path = pathlib.Path(__file__).parent / "data" / "igrf12coeffs.txt"
        with tempfile.TemporaryDirectory() as outpath:
            cli.main("import-igrf", "-o", outpath, str(path))

        self.import_igrf_mock.assert_called_once_with(str(path))
        self.wmmdata_mock.save.assert_called_once_with(outpath, False)



"""Internal utilities."""

import numpy as np


def as_contiguous_1d_llh(lat, lon, h, dtype=np.float64):
    """Return lat, lon and h as contiguous arrays of the specified data type.

    Shape consistency is checked and, in the case `h` is provided as scalar,
    a new array having the proper shape and `h` value is created and returned.

    :returns:
        lat, lon, h, shape
    """
    is_scalar = np.isscalar(lat)

    lat = np.asarray(lat)
    lon = np.asarray(lon)
    h = np.asarray(h)

    for name, param in (("lat", lat), ("lon", lon), ("h", h)):
        dt = param.dtype
        if not (
            np.issubdtype(dt, np.floating) or np.issubdtype(dt, np.integer)
        ):
            raise TypeError(f"{name}: {param!r}")

    shape = lat.shape
    if lon.shape != shape or (h.size > 1 and h.shape != shape):
        raise ValueError("lat, lon and h shall have the same shape")

    size = lat.size

    lat = np.ascontiguousarray(lat.reshape([size]), dtype=dtype)
    lon = np.ascontiguousarray(lon.reshape([size]), dtype=dtype)
    if h.size > 1:
        h = np.ascontiguousarray(h.reshape([size]), dtype=dtype)
    else:
        h = np.full([size], h.item(), dtype)

    shape = shape if not is_scalar else None

    return lat, lon, h, shape


def as_contiguous_1d_components(*args, labels=None, dtype=np.float64):
    """Return the inputs as contiguous arrays of the specified data type."""
    if labels is None:
        labels = [f"component_{i}" for i in range(len(args))]

    is_scalar = np.isscalar(args[0])

    arglist = [np.asarray(item) for item in args]

    shape: tuple | None = arglist[0].shape
    for name, param in zip(labels, arglist):
        dt = param.dtype
        if not (
            np.issubdtype(dt, np.floating) or np.issubdtype(dt, np.integer)
        ):
            raise TypeError(f"{name}: {param!r}")

        if param.shape != shape:
            raise ValueError("not all components have the same shape")

    size = arglist[0].size
    for i in range(len(arglist)):
        arglist[i] = np.ascontiguousarray(
            arglist[i].reshape([size]), dtype=dtype
        )

    shape = shape if not is_scalar else None

    return arglist + [shape]


def reshape_components(shape, *args) -> np.ndarray | list[np.ndarray]:
    """Reshape all args to the specified shape."""
    is_scalar = True if shape is None else False

    if is_scalar:
        argslist = [item.item() for item in args]
    else:
        argslist = [item.reshape(shape) for item in args]

    if len(argslist) == 1:
        out = argslist[0]
    else:
        out = argslist

    return out



# -*- coding: utf-8 -*-
# cython: language_level=3
# distutils: language=c++

import os

import numpy as np

cimport cython
from libcpp.string cimport string

from .magnetic cimport CMagneticModel

from .error import GeographicErr
from ._utils import (
    as_contiguous_1d_llh,
    as_contiguous_1d_components,
    reshape_components,
)
from ._typing import PathType


cdef class MagneticFieldModel:
    """Model of the earth's magnetic field.

    Evaluate the earth's magnetic field according to a model.  At present only
    internal magnetic fields are handled.  These are due to the earth's code
    and crust; these vary slowly (over many years).  Excluded are the effects
    of currents in the ionosphere and magnetosphere which have daily and
    annual variations.

    See https://geographiclib.sourceforge.io/html/magnetic.html for details
    of how to install the magnetic models and the data format.

    :param str name:
        the name of the model
    :param path:
        (optional) directory for data file
    :raises GeographicErr:
         if the data file cannot be found, is unreadable, or is corrupt
    :raises MemoryError:
         if the memory necessary for storing the model can't be allocated

    A filename is formed by appending ".wmm" (World Magnetic Model) to
    the name.  If path is specified (and is non-empty), then the file
    is loaded from directory, path.  Otherwise the path is given by the
    :meth:`MagneticFieldModel.default_magnetic_path`.

    This file contains the metadata which specifies the properties of the
    model.  The coefficients for the spherical harmonic sums are obtained
    from a file obtained by appending ".cof" to metadata file (so the
    filename ends in ".wwm.cof").

    The model is not tied to a particular ellipsoidal model of the earth.
    The final earth argument to the constructor specifies an ellipsoid to
    allow geodetic coordinates to the transformed into the spherical
    coordinates used in the spherical harmonic sum.

    Example
    -------
    ::

        # Example of using the `geomodels.MagneticModel` class.
        # This requires that the wmm2010 magnetic model be installed; see
        # https://geographiclib.sourceforge.io/C++/doc/magnetic.html#magneticinst

        from geomodels import MagneticModel

        mag MagneticModel("wmm2010")

        # Mt Everest
        lat = 27.99
        lon = 86.93
        h = 8820
        t = 2012

        Bx, By, Bz = mag(t, lat,lon, h, Bx, By, Bz)
        H, F, D, I = MagneticModel.compute_field_components(Bx, By, Bz)

        print(f"horizontal magnetic field (nT): {H}")
        print(f"total magnetic field (nT): {F}")
        print(f"declination of the field (degrees east of north): {D}")
        print(f"inclination of the field (degrees down from horizontal): {I}")
    """
    cdef CMagneticModel *_ptr

    def __cinit__(self, name: str | None = None, path: PathType = ""):
        # @TODO: support for 'earth' parameter (default: WGS84)
        cdef string c_name = CMagneticModel.DefaultMagneticName()
        cdef string c_path = os.fsencode(os.fspath(path))

        if name is not None:
            c_name = os.fsencode(name)

        try:
            with nogil:
                self._ptr = new CMagneticModel(c_name, c_path)
        except RuntimeError as exc:
            raise GeographicErr(str(exc)) from exc

    def __dealloc__(self):
        del self._ptr

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _compute(
        self, double t, double[::1] vlat, double[::1] vlon, double[::1] vh
    ):
        cdef long size = vlat.size
        dtype = np.float64

        Bx = np.empty(shape=[size], dtype=dtype)
        By = np.empty(shape=[size], dtype=dtype)
        Bz = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vBx = Bx
        cdef double[::1] vBy = By
        cdef double[::1] vBz = Bz

        cdef long i = 0
        with nogil:
            for i in range(size):
                cython.operator.dereference(self._ptr)(
                    t, vlat[i], vlon[i], vh[i], vBx[i], vBy[i], vBz[i]
                )

        return Bx, By, Bz

    def __call__(self, double t, lat, lon, h):
        """Compute the magnetic field.

        Evaluate the components of the geomagnetic field.

        :param t:
            the time (years)
        :param lat:
            latitude of the point (degrees)
        :param lon:
            longitude of the point (degrees)
        :param h:
            the height of the point above the ellipsoid (meters)
        :returns:
            Bx, By, Bz: the easterly, northerly and vertical (up)
            components of the magnetic field (nT).
        """
        dtype = np.float64
        lat, lon, h, shape = as_contiguous_1d_llh(lat, lon, h, dtype)
        Bx, By, Bz = self._compute(t, lat, lon, h)
        return reshape_components(shape, Bx, By, Bz)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _compute_with_rate(
        self, double t, double[::1] vlat, double[::1] vlon, double[::1] vh
    ):
        cdef long size = vlat.size
        dtype = np.float64

        Bx = np.empty(shape=[size], dtype=dtype)
        By = np.empty(shape=[size], dtype=dtype)
        Bz = np.empty(shape=[size], dtype=dtype)
        Bxt = np.empty(shape=[size], dtype=dtype)
        Byt = np.empty(shape=[size], dtype=dtype)
        Bzt = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vBx = Bx
        cdef double[::1] vBy = By
        cdef double[::1] vBz = Bz
        cdef double[::1] vBxt = Bxt
        cdef double[::1] vByt = Byt
        cdef double[::1] vBzt = Bzt

        cdef long i = 0
        with nogil:
            for i in range(size):
                cython.operator.dereference(self._ptr)(
                    t,
                    vlat[i],
                    vlon[i],
                    vh[i],
                    vBx[i],
                    vBy[i],
                    vBz[i],
                    vBxt[i],
                    vByt[i],
                    vBzt[i],
                )

        return Bx, By, Bz, Bxt, Byt, Bzt

    def compute_with_rate(self, double t, lat, lon, h):
        """Compute the magnetic field and its rate.

        Evaluate the components of the geomagnetic field and
        their time derivatives

        :param t:
            the time (years)
        :param lat:
            latitude of the point (degrees)
        :param lon:
            longitude of the point (degrees)
        :param h:
            the height of the point above the ellipsoid (meters)
        :returns:
            `Bx`, `By`, `Bz`, `Bxt`, `Byt`, `Bzt`: where
            `Bx`, `By`, `Bz` are the easterly, northerly and vertical (up)
            components of the magnetic field (nT) respectively, and
            `Bxt`, `Byt`, `Bzt` are the rate (nT/yr) of change of
            `Bx`, `By` and `Bz` respectively.
        """
        dtype = np.float64
        lat, lon, h, shape = as_contiguous_1d_llh(lat, lon, h, dtype)
        Bx, By, Bz, Bxt, Byt, Bzt = self._compute_with_rate(t, lat, lon, h)
        return reshape_components(shape, Bx, By, Bz, Bxt, Byt, Bzt)

    # @TODO: MagneticCircle Circle(real t, real lat, real h) const
    # def circle(...):
    #     """Create a MagneticCircle object.
    #
    #     A MagneticCircle object to allow the geomagnetic field at many
    #     points with constant \e lat, \e h, and \e t and varying `lon`
    #     to be computed efficiently.
    #
    #     :param t:
    #         the time (years).
    #     :param lat:
    #         latitude of the point (degrees).
    #     :param h:
    #         the height of the point above the ellipsoid (meters).
    #     :raises MenoryError:
    #         if the memory necessary for creating a MagneticCircle
    #         can't be allocated.
    #     :returns:
    #         a MagneticCircle object whose MagneticCircle.__call__(lon)
    #         member function computes the field at particular values of
    #         `lon`.
    #
    #     If the field at several points on a circle of latitude need to be
    #     calculated then creating a MagneticCircle and using its member
    #     functions will be substantially faster, especially for high-degree
    #     models.
    #     """
    #     pass

    @staticmethod
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef _field_components(
        double[::1] vBx, double[::1] vBy, double[::1] vBz
    ):
        cdef long size = vBx.size
        dtype = np.float64

        H = np.empty(shape=[size], dtype=dtype)
        F = np.empty(shape=[size], dtype=dtype)
        D = np.empty(shape=[size], dtype=dtype)
        I = np.empty(shape=[size], dtype=dtype)

        cdef double[::1] vH = H
        cdef double[::1] vF = F
        cdef double[::1] vD = D
        cdef double[::1] vI = I

        cdef long i = 0
        with nogil:
            for i in range(size):
                CMagneticModel.FieldComponents(
                    vBx[i], vBy[i], vBz[i], vH[i], vF[i], vD[i], vI[i]
                )

        return H, F, D, I

    @staticmethod
    def field_components(Bx, By, Bz):
        """Compute various quantities dependent on the magnetic field.

        :param float Bx:
            the x (easterly) component of the magnetic field (nT)
        :param float By:
            the y (northerly) component of the magnetic field (nT)
        :param float Bz:
            the z (vertical, up positive) component of the magnetic field (nT)
        :returns:
            * H the horizontal magnetic field (nT)
            * F the total magnetic field (nT)
            * D the declination of the field (degrees east of north)
            * I the inclination of the field (degrees down from horizontal)
        """
        dtype = np.float64
        Bx, By, Bz, shape = as_contiguous_1d_components(
            Bx, By, Bz, labels=['Bx', 'By', 'Bz'], dtype=dtype
        )
        H, F, D, I = MagneticFieldModel._field_components(Bx, By, Bz)
        return reshape_components(shape, H, F, D, I)

    # @staticmethod
    # def field_components_and_rate(
    #     double Bx, double By, double Bz, double Bxt, double Byt, double Bzt
    # ):
    #     """Compute quantities dependent on the magnetic field and its rate
    #     of change.
    #
    #     :param float Bx:
    #         the x (easterly) component of the magnetic field (nT)
    #     :param float By:
    #         the y (northerly) component of the magnetic field (nT)
    #     :param float Bz:
    #         the z (vertical, up positive) component of the magnetic field (nT)
    #     :param float Bxt:
    #         the rate of change of Bx (nT/yr)
    #     :param float Byt:
    #         the rate of change of By (nT/yr)
    #     :param float Bzt:
    #         the rate of change of Bz (nT/yr)
    #
    #     :returns:
    #         * H the horizontal magnetic field (nT)
    #         * F the total magnetic field (nT)
    #         * D the declination of the field (degrees east of north)
    #         * I the inclination of the field (degrees down from horizontal)
    #         * Ht the rate of change of H (nT/yr)
    #         * Ft the rate of change of F (nT/yr)
    #         * Dt the rate of change of D (degrees/yr)
    #         * It the rate of change of I (degrees/yr)
    #     """
    #     cdef double H=0, F=0, D=0, I=0, Ht=0, Ft=0, Dt=0, It=0
    #     with nogil:
    #         MagneticModel.FieldComponents(
    #             Bx, By, Bz, Bxt, Byt, Bzt, H, F, D, I, Ht, Ft, Dt, It
    #         )
    #     return H, F, D, I, Ht, Ft, Dt, It

    def description(self) -> str:
        """Return the description of the magnetic model.

        Return the description of the magnetic model if available,
        from the :meth:`MagneticFieldModel.description` file in the
        data file; if absent, return "NONE".
        """
        return self._ptr.Description().decode('utf-8')

    def datetime(self) -> str:
        """Return date of the model.

        Return date of the model, if available, from the ReleaseDate
        field in the data file; if absent, return "UNKNOWN".
        """
        return self._ptr.DateTime().decode('utf-8')

    def magnetic_file(self) -> str:
        """Full file name used to load the magnetic model."""
        return self._ptr.MagneticFile().decode('utf-8')

    def magnetic_model_name(self) -> str:
        """Name used to load the magnetic model.

        The 'name' used to load the magnetic model (from the first
        argument of the constructor, but this may be overridden by
        the model file).
        """
        return self._ptr.MagneticModelName().decode('utf-8')

    def magnetic_model_directory(self) -> str:
        """Directory used to load the magnetic model."""
        return self._ptr.MagneticModelDirectory().decode('utf-8')

    def min_height(self) -> float:
        """Minimum height.

        The minimum height above the ellipsoid (in meters) for which
        this MagneticModel should be used.

        Because the model will typically provide useful results
        slightly outside the range of allowed heights,
        no check of t argument is made by the
        :meth:`MagneticFieldModel.__call__` operator.
        """
        return self._ptr.MinHeight()

    def max_height(self) -> float:
        """Maximum height.

        The maximum height above the ellipsoid (in meters) for which
        this MagneticModel should be used.

        Because the model will typically provide useful results
        slightly outside the range of allowed heights,
        no check of t argument is made by the
        :meth:`MagneticFieldModel.__call__` operator.
        """
        return self._ptr.MaxHeight()

    def min_time(self) -> float:
        """The minimum time (in years) for which this model should be used.

        Because the model will typically provide useful results
        slightly outside the range of allowed times,
        no check of t argument is made by the
        :meth:`MagneticFieldModel.__call__` operator.
        """
        return self._ptr.MinTime()

    def max_time(self) -> float:
        """The maximum time (in years) for which this model should be used.

        Because the model will typically provide useful results
        slightly outside the range of allowed times,
        no check of the `t` argument is made by the
        :meth:`MagneticFieldModel.__call__` operator.
        """
        return self._ptr.MaxTime()

    def equatorial_radius(self) -> float:
        """The equatorial radius of the ellipsoid (meters).

        This is the value of 'a' inherited from the Geocentric object
        used in the constructor.
        """
        return self._ptr.EquatorialRadius()

    def flattening(self) -> float:
        """The flattening of the ellipsoid.

        This is the value inherited from the Geocentric object
        used in the constructor.
        """
        return self._ptr.Flattening()

    def degree(self) -> int:
        """The maximum degree of the components of the model (`Nmax`)."""
        return self._ptr.Degree()

    def order(self) -> int:
        """The maximum order of the components of the model (`Mmax`)."""
        return self._ptr.Order()

    @staticmethod
    def default_magnetic_path() -> str:
        """Return the default path for magnetic model data files.

        This is the value of the environment variable
        `GEOGRAPHICLIB_MAGNETIC_PATH`, if set; otherwise, it is
        `$GEOGRAPHICLIB_DATA/magnetic` if the environment variable
        `GEOGRAPHICLIB_DATA` is set; otherwise, it is a compile-time
        default.
        """
        return CMagneticModel.DefaultMagneticPath().decode('utf-8')

    @staticmethod
    def default_magnetic_name() -> str:
        """The default name for the magnetic model.

        This is the value of the environment variable
        `GEOGRAPHICLIB_MAGNETIC_NAME`, if set; otherwise,
        it is "wmm2015".
        The :class:`MagneticFieldModel` class does not use this function;
        it is just provided as a convenience for a calling program
        when constructing a :class:`MagneticFieldModel` object.
        """
        return CMagneticModel.DefaultMagneticName().decode('utf-8')



import os
import shutil
import pathlib
import datetime
import tempfile
import unittest

import numpy as np
from numpy import testing as npt

from geomodels import MagneticFieldModel, get_default_data_path
from geomodels.tests.utils import dms_to_dec


class StaticMethodsTestCase(unittest.TestCase):
    def test_default_magnetic_path(self):
        self.assertIsInstance(MagneticFieldModel.default_magnetic_path(), str)
        self.assertEqual(
            pathlib.Path(MagneticFieldModel.default_magnetic_path()),
            pathlib.Path(get_default_data_path()).joinpath("magnetic"),
        )

    def test_default_magnetic_name(self):
        names = (
            "wmm2010",
            "wmm2015",
            "wmm2015v2",
            "wmm2020",
            "wmm2025",
            "wmmhr2025",
            "igrf11",
            "igrf12",
            "igrf13",
            "igrf14",
            "emm2010",
            "emm2015",
            "emm2017",
        )
        self.assertTrue(MagneticFieldModel.default_magnetic_name() in names)


class InstantiationTestCase00(unittest.TestCase):
    def test_no_args(self):
        model = MagneticFieldModel()
        self.assertIsInstance(model, MagneticFieldModel)
        self.assertEqual(
            model.magnetic_model_name(),
            MagneticFieldModel.default_magnetic_name(),
        )


class InstantiationTestCase01(unittest.TestCase):
    MODEL_NAME = "wmm2015"

    def test_name(self):
        model = MagneticFieldModel(self.MODEL_NAME)
        self.assertIsInstance(model, MagneticFieldModel)
        self.assertEqual(model.magnetic_model_name(), self.MODEL_NAME)

    def test_default_path(self):
        path = MagneticFieldModel.default_magnetic_path()
        model = MagneticFieldModel(self.MODEL_NAME, path)
        self.assertEqual(model.magnetic_model_name(), self.MODEL_NAME)
        self.assertEqual(model.magnetic_model_directory(), path)

    def test_custom_path(self):
        default_path = MagneticFieldModel.default_magnetic_path()
        with tempfile.TemporaryDirectory() as dirname:
            magnetic_path = os.path.join(
                dirname, os.path.basename(default_path)
            )
            shutil.copytree(default_path, magnetic_path)
            model = MagneticFieldModel(self.MODEL_NAME, magnetic_path)
            self.assertEqual(model.magnetic_model_name(), self.MODEL_NAME)
            self.assertEqual(model.magnetic_model_directory(), magnetic_path)

    def test_custom_path_from_env01(self):
        default_path = pathlib.Path(MagneticFieldModel.default_magnetic_path())
        with tempfile.TemporaryDirectory() as dirname:
            magnetic_path = pathlib.Path(dirname) / default_path.name
            shutil.copytree(default_path, magnetic_path)

            old_env = os.environ.get("GEOGRAPHICLIB_DATA")
            os.environ["GEOGRAPHICLIB_DATA"] = dirname
            try:
                model = MagneticFieldModel(self.MODEL_NAME)
                self.assertEqual(model.magnetic_model_name(), self.MODEL_NAME)
                self.assertEqual(
                    pathlib.Path(model.magnetic_model_directory()),
                    magnetic_path,
                )
            finally:
                if old_env is None:
                    del os.environ["GEOGRAPHICLIB_DATA"]
                else:
                    os.environ["GEOGRAPHICLIB_DATA"] = old_env

    def test_custom_path_from_env02(self):
        default_path = MagneticFieldModel.default_magnetic_path()
        with tempfile.TemporaryDirectory() as dirname:
            magnetic_path = os.path.join(
                dirname, os.path.basename(default_path)
            )
            shutil.copytree(default_path, magnetic_path)

            old_env = os.environ.get("GEOGRAPHICLIB_MAGNETIC_PATH")
            os.environ["GEOGRAPHICLIB_MAGNETIC_PATH"] = magnetic_path
            try:
                model = MagneticFieldModel(self.MODEL_NAME)
                self.assertEqual(model.magnetic_model_name(), self.MODEL_NAME)
                self.assertEqual(
                    model.magnetic_model_directory(), magnetic_path
                )
            finally:
                if old_env is None:
                    del os.environ["GEOGRAPHICLIB_MAGNETIC_PATH"]
                else:
                    os.environ["GEOGRAPHICLIB_MAGNETIC_PATH"] = old_env


class InstantiationTestCase02(InstantiationTestCase01):
    MODEL_NAME = "igrf12"


class InfoMethodsTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.name = MagneticFieldModel.default_magnetic_name()
        self.datapath = MagneticFieldModel.default_magnetic_path()
        self.model = MagneticFieldModel(self.name, self.datapath)

    def test_description(self):
        description = self.model.description()
        self.assertIsInstance(description, str)
        self.assertNotEqual(description, "NONE")

    def test_datetime(self):
        datestr = self.model.datetime()
        self.assertIsInstance(datestr, str)
        self.assertNotEqual(datestr, "UNKNOWN")
        date = datetime.datetime.strptime(datestr, "%Y-%m-%d")
        # date = datetime.datetime.strptime(datestr, '%Y-%m-%d %H:%M:%S')
        self.assertLess(date, datetime.datetime.now())

    def test_magnetic_file(self):
        filename = self.model.magnetic_file()
        self.assertIn(self.name, filename)
        self.assertIn(self.datapath, filename)

    def test_magnetic_model_name(self):
        name = self.model.magnetic_model_name()
        self.assertEqual(name, self.name)

    def test_magnetic_model_directory(self):
        path = self.model.magnetic_model_directory()
        self.assertEqual(path, self.datapath)

    def test_min_height(self):
        self.assertIsInstance(self.model.min_height(), float)

    def test_max_height(self):
        self.assertIsInstance(self.model.max_height(), float)
        self.assertGreater(self.model.max_height(), 0)

    def test_min_max_height(self):
        self.assertLess(self.model.min_height(), self.model.max_height())

    def test_min_time(self):
        self.assertIsInstance(self.model.min_time(), float)
        min_times = {
            "emm2010": 2010,
            "emm2015": 2000,
            "emm2017": 2000,
            "igrf11": 1900,
            "igrf12": 1900,
            "igrf13": 1900,
            "igrf14": 1900,
            "wmm2010": 2010,
            "wmm2015v2": 2015,
            "wmm2015": 2015,
            "wmm2020": 2020,
            "wmm2025": 2025,
            "wmmhr2025": 2025,
        }
        if self.name in min_times:
            self.assertEqual(self.model.min_time(), min_times[self.name])
        else:
            self.assertLess(self.model.min_time(), datetime.date.today().year)

    def test_max_time(self):
        self.assertIsInstance(self.model.max_time(), float)
        max_times = {
            "emm2010": 2015,
            "emm2015": 2020,
            "emm2017": 2022,
            "igrf11": 2015,
            "igrf12": 2020,
            "igrf13": 2025,
            "igrf14": 2030,
            "wmm2010": 2015,
            "wmm2015v2": 2020,
            "wmm2015": 2020,
            "wmm2020": 2025,
            "wmm2025": 2030,
            "wmmhr2025": 2030,
        }
        if self.name in max_times:
            self.assertEqual(self.model.max_time(), max_times[self.name])
        else:
            self.assertGreater(
                self.model.max_time(), datetime.date.today().year
            )

    def test_min_max_time(self):
        self.assertLess(self.model.min_time(), self.model.max_time())

    def test_major_radius(self):
        self.assertIsInstance(self.model.equatorial_radius(), float)
        self.assertGreater(self.model.equatorial_radius(), 5e6)

    def test_flattening(self):
        self.assertIsInstance(self.model.flattening(), float)
        self.assertGreater(self.model.flattening(), 0)
        self.assertLess(self.model.flattening(), 1)

    def test_degree(self):
        self.assertIsInstance(self.model.degree(), int)
        self.assertGreaterEqual(self.model.degree(), 0)
        self.assertLessEqual(self.model.degree(), 360)

    def test_order(self):
        self.assertIsInstance(self.model.order(), int)
        self.assertGreater(self.model.order(), 0)
        self.assertLessEqual(self.model.order(), 12)


class ComputationTestCase(unittest.TestCase):
    # MODEL_NAME = MagneticFieldModel.default_magnetic_name()
    MODEL_NAME = "wmm2015"

    YEAR = 2016.0  # 2016-01-01
    LAT = +dms_to_dec(16, 46, 33)  # 16:46:33N
    LON = -dms_to_dec(3, 0, 34)  # 03:00:34W
    HEIGHT = +300  # [m]

    BX = -1251.3634113  # NOTE: east component (5th item)
    BY = +33848.7341446  # NOTE: north component (4th item)
    BZ = -7293.8535382  # NOTE: inverted sign
    BXT = +53.6988656  # NOTE: east component (5th item, second line)
    BYT = +33.7765829  # NOTE: north component (4th item, second line)
    BZT = +41.3769946  # NOTE: inverted sign
    H = +33871.8572503
    F = +34648.2757582
    D = -2.11721965
    I = +12.15231735  # noqa: E741

    # $ echo 2016-01-01 16:46:33N 3:00:34W 300 | MagneticField -r -p 7
    # -2.11721965 12.15231735 33871.8572503 33848.7341446 \
    # -1251.3634113 7293.8535382 34648.2757582
    # 0.09288284 -0.07794874 31.7696715 33.7765829 \
    # 53.6988656 -41.3769946 22.3474336
    #
    # $ echo 2016.0 16.775833333333335 -3.0094444444444446 300 | \
    # MagneticField -r -p 7
    # -2.11721965 12.15231735 33871.8572503 33848.7341446 \
    # -1251.3634113 7293.8535382 34648.2757582
    # 0.09288284 -0.07794874 31.7696715 33.7765829 \
    # 53.6988656 -41.3769946 22.3474336

    def setUp(self) -> None:
        self.model = MagneticFieldModel(self.MODEL_NAME)

    def test_field_components(self):
        H, F, D, I = MagneticFieldModel.field_components(  # noqa: E741,N806
            self.BX, self.BY, self.BZ
        )
        npt.assert_allclose([H, F, D, I], [self.H, self.F, self.D, self.I])

    # @TODO
    # def test_field_components_and_rate(self):
    #     pass

    # @TODO
    # def test_circle(self):
    #     pass

    def test_compute_scalar(self):
        fld_x, fld_y, fld_z = self.model(
            self.YEAR, self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose([fld_x, fld_y, fld_z], [self.BX, self.BY, self.BZ])

    def test_compute_scalar_with_rate(self):
        (fld_x, fld_y, fld_z, fld_xt, fld_yt, fld_zt) = (
            self.model.compute_with_rate(
                self.YEAR, self.LAT, self.LON, self.HEIGHT
            )
        )
        npt.assert_allclose([fld_x, fld_y, fld_z], [self.BX, self.BY, self.BZ])
        npt.assert_allclose(
            [fld_xt, fld_yt, fld_zt], [self.BXT, self.BYT, self.BZT]
        )


class VectorComputationTestCase(unittest.TestCase):
    # MODEL_NAME = MagneticFieldModel.default_magnetic_name()
    MODEL_NAME = "wmm2015"

    YEAR = 2016.0
    LAT = np.asarray(
        [
            [+dms_to_dec(16, 46, 33), -dms_to_dec(16, 46, 43)],
            [-dms_to_dec(16, 56, 33), +dms_to_dec(16, 56, 43)],
        ]
    )
    LON = np.asarray(
        [
            [-dms_to_dec(3, 0, 34), +dms_to_dec(3, 0, 44)],
            [+dms_to_dec(3, 10, 34), -dms_to_dec(3, 10, 44)],
        ]
    )
    HEIGHT = np.asarray(
        [
            [+300, +400000],
            [+400000, +300],
        ]
    )
    BX = np.asarray(
        [
            [-1251.3634113, -2917.6087863],
            [-2908.0688421, -1262.5153146],
        ]
    )
    BY = np.asarray(
        [
            [+33848.7341446, +13455.1508250],
            [+13378.9665352, +33836.6298435],
        ]
    )
    BZ = np.asarray(
        [
            [-7293.85353820, +20132.4047379],
            [+20169.4124889, -7539.67870600],
        ]
    )
    BXT = np.asarray(
        [
            [+53.6988656, +48.1739432],
            [+48.0353017, +53.7715374],
        ]
    )
    BYT = np.asarray(
        [
            [+33.7765829, -37.4184172],
            [-37.5102493, +33.9119117],
        ]
    )
    BZT = np.asarray(
        [
            [+41.3769946, +28.8842256],
            [+27.8740361, +41.1851008],
        ]
    )
    H = np.asarray(
        [
            [+33871.8572503, +13767.8438673],
            [+13691.3699073, +33860.1751928],
        ]
    )
    F = np.asarray(
        [
            [+34648.2757582, +24389.9004772],
            [+24377.4241889, +34689.4540037],
        ]
    )
    D = np.asarray(
        [
            [-2.117219650, -12.23458341],
            [-12.26312924, -2.136833910],
        ]
    )
    I = np.asarray(  # noqa: E741
        [
            [+12.15231735, -55.63314796],
            [-55.83061445, +12.55330784],
        ]
    )

    def setUp(self) -> None:
        self.model = MagneticFieldModel(self.MODEL_NAME)

    def test_field_components_vector(self):
        H, F, D, I = MagneticFieldModel.field_components(  # noqa: E741,N806
            self.BX.flatten(), self.BY.flatten(), self.BZ.flatten()
        )
        npt.assert_allclose(H, self.H.flatten())
        npt.assert_allclose(F, self.F.flatten())
        npt.assert_allclose(D, self.D.flatten())
        npt.assert_allclose(I, self.I.flatten())

    def test_field_components_matrix(self):
        H, F, D, I = MagneticFieldModel.field_components(  # noqa: E741,N806
            self.BX, self.BY, self.BZ
        )
        npt.assert_allclose(H, self.H)
        npt.assert_allclose(F, self.F)
        npt.assert_allclose(D, self.D)
        npt.assert_allclose(I, self.I)

    # @TODO
    # def test_field_components_and_rate_xxx(self):
    #     pass

    # @TODO
    # def test_circle_xxx(self):
    #     pass

    def test_compute_vector(self):
        fld_x, fld_y, fld_z = self.model(
            self.YEAR,
            self.LAT.flatten(),
            self.LON.flatten(),
            self.HEIGHT.flatten(),
        )
        npt.assert_allclose(fld_x, self.BX.flatten())
        npt.assert_allclose(fld_y, self.BY.flatten())
        npt.assert_allclose(fld_z, self.BZ.flatten())

    def test_compute_vector_with_rate(self):
        (fld_x, fld_y, fld_z, fld_xt, fld_yt, fld_zt) = (
            self.model.compute_with_rate(
                self.YEAR,
                self.LAT.flatten(),
                self.LON.flatten(),
                self.HEIGHT.flatten(),
            )
        )
        npt.assert_allclose(fld_x, self.BX.flatten())
        npt.assert_allclose(fld_y, self.BY.flatten())
        npt.assert_allclose(fld_z, self.BZ.flatten())
        npt.assert_allclose(fld_xt, self.BXT.flatten())
        npt.assert_allclose(fld_yt, self.BYT.flatten())
        npt.assert_allclose(fld_zt, self.BZT.flatten())

    def test_compute_matrix(self):
        fld_x, fld_y, fld_z = self.model(
            self.YEAR, self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose(fld_x, self.BX)
        npt.assert_allclose(fld_y, self.BY)
        npt.assert_allclose(fld_z, self.BZ)

    def test_compute_matrix_with_rate(self):
        (fld_x, fld_y, fld_z, fld_xt, fld_yt, fld_zt) = (
            self.model.compute_with_rate(
                self.YEAR, self.LAT, self.LON, self.HEIGHT
            )
        )
        npt.assert_allclose(fld_x, self.BX)
        npt.assert_allclose(fld_y, self.BY)
        npt.assert_allclose(fld_z, self.BZ)
        npt.assert_allclose(fld_xt, self.BXT)
        npt.assert_allclose(fld_yt, self.BYT)
        npt.assert_allclose(fld_zt, self.BZT)


class ConstHeightVectorComputationTestCase(unittest.TestCase):
    # MODEL_NAME = MagneticFieldModel.default_magnetic_name()
    MODEL_NAME = "wmm2015"

    YEAR = 2016.0
    LAT = np.asarray(
        [
            [+dms_to_dec(16, 46, 33), +dms_to_dec(16, 56, 43)],
            [+dms_to_dec(16, 46, 33), +dms_to_dec(16, 56, 43)],
        ]
    )
    LON = np.asarray(
        [
            [-dms_to_dec(3, 0, 34), -dms_to_dec(3, 10, 44)],
            [-dms_to_dec(3, 0, 34), -dms_to_dec(3, 10, 44)],
        ]
    )
    HEIGHT = +300
    BX = np.asarray(
        [
            [-1251.3634113, -1262.5153146],
            [-1251.3634113, -1262.5153146],
        ]
    )
    BY = np.asarray(
        [
            [+33848.7341446, +33836.6298435],
            [+33848.7341446, +33836.6298435],
        ]
    )
    BZ = np.asarray(
        [
            [-7293.85353820, -7539.67870600],
            [-7293.85353820, -7539.67870600],
        ]
    )
    BXT = np.asarray(
        [
            [+53.6988656, +53.7715374],
            [+53.6988656, +53.7715374],
        ]
    )
    BYT = np.asarray(
        [
            [+33.7765829, +33.9119117],
            [+33.7765829, +33.9119117],
        ]
    )
    BZT = np.asarray(
        [
            [+41.3769946, +41.1851008],
            [+41.3769946, +41.1851008],
        ]
    )
    H = np.asarray(
        [
            [+33871.8572503, +33860.1751928],
            [+33871.8572503, +33860.1751928],
        ]
    )
    F = np.asarray(
        [
            [+34648.2757582, +34689.4540037],
            [+34648.2757582, +34689.4540037],
        ]
    )
    D = np.asarray(
        [
            [-2.11721965, -2.13683391],
            [-2.11721965, -2.13683391],
        ]
    )
    I = np.asarray(  # noqa: E741
        [
            [+12.15231735, +12.55330784],
            [+12.15231735, +12.55330784],
        ]
    )

    def setUp(self) -> None:
        self.model = MagneticFieldModel(self.MODEL_NAME)

    # @TODO
    # def test_circle_xxx(self):
    #     pass

    def test_compute_vector(self):
        fld_x, fld_y, fld_z = self.model(
            self.YEAR, self.LAT.flatten(), self.LON.flatten(), self.HEIGHT
        )
        npt.assert_allclose(fld_x, self.BX.flatten())
        npt.assert_allclose(fld_y, self.BY.flatten())
        npt.assert_allclose(fld_z, self.BZ.flatten())

    def test_compute_vector_with_rate(self):
        (fld_x, fld_y, fld_z, fld_xt, fld_yt, fld_zt) = (
            self.model.compute_with_rate(
                self.YEAR,
                self.LAT.flatten(),
                self.LON.flatten(),
                self.HEIGHT,
            )
        )
        npt.assert_allclose(fld_x, self.BX.flatten())
        npt.assert_allclose(fld_y, self.BY.flatten())
        npt.assert_allclose(fld_z, self.BZ.flatten())
        npt.assert_allclose(fld_xt, self.BXT.flatten())
        npt.assert_allclose(fld_yt, self.BYT.flatten())
        npt.assert_allclose(fld_zt, self.BZT.flatten())

    def test_compute_matrix(self):
        fld_x, fld_y, fld_z = self.model(
            self.YEAR, self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose(fld_x, self.BX)
        npt.assert_allclose(fld_y, self.BY)
        npt.assert_allclose(fld_z, self.BZ)

    def test_compute_matrix_with_rate(self):
        (fld_x, fld_y, fld_z, fld_xt, fld_yt, fld_zt) = (
            self.model.compute_with_rate(
                self.YEAR, self.LAT, self.LON, self.HEIGHT
            )
        )
        npt.assert_allclose(fld_x, self.BX)
        npt.assert_allclose(fld_y, self.BY)
        npt.assert_allclose(fld_z, self.BZ)
        npt.assert_allclose(fld_xt, self.BXT)
        npt.assert_allclose(fld_yt, self.BYT)
        npt.assert_allclose(fld_zt, self.BZT)



"""Geographic data models."""

from ._ext import (  # noqa: F401
    lib_version_info,
    lib_version_str,
    GeoidModel,
    GravityModel,
    EHeightConvDir,
    MagneticFieldModel,
)
from .data import (  # noqa: F401
    EModelGroup,
    EModelType,
    EGeoidModel,
    EGravityModel,
    EMagneticModel,
    EArchiveType,
    get_default_data_path,
    get_model_url,
    install,
)
from .wmmf import MetaData, SphCoeffSet, WmmData, import_igrf_txt  # noqa: F401
from .error import GeographicErr  # noqa: F401
from .tests import test  # noqa: F401
from ._version import __version__  # noqa: F401



import os
import shutil
import pathlib
import datetime
import tempfile
import unittest

import numpy as np
from numpy import testing as npt

from geomodels import GeoidModel, EHeightConvDir, get_default_data_path
from geomodels.tests.utils import dms_to_dec


class StaticMethodsTestCase(unittest.TestCase):
    def test_default_geoid_path(self):
        self.assertIsInstance(GeoidModel.default_geoid_path(), str)
        self.assertEqual(
            pathlib.Path(GeoidModel.default_geoid_path()),
            pathlib.Path(get_default_data_path()).joinpath("geoids"),
        )

    def test_default_geoid_name(self):
        names = (
            "egm84-30",
            "egm84-15",
            "egm96-15",
            "egm96-5",
            "egm2008-5",
            "egm2008-2_5",
            "egm2008-1",
        )
        self.assertTrue(GeoidModel.default_geoid_name() in names)


class InstantiationTestCase00(unittest.TestCase):
    def test_no_args(self):
        model = GeoidModel()
        self.assertIsInstance(model, GeoidModel)
        self.assertEqual(model.geoid_name(), GeoidModel.default_geoid_name())


class InstantiationTestCase(unittest.TestCase):
    MODEL_NAME = "egm96-5"

    def test_name(self):
        model = GeoidModel(self.MODEL_NAME)
        self.assertIsInstance(model, GeoidModel)
        self.assertEqual(model.geoid_name(), self.MODEL_NAME)

    def test_default_path(self):
        path = GeoidModel.default_geoid_path()
        model = GeoidModel(self.MODEL_NAME, path)
        self.assertEqual(model.geoid_name(), self.MODEL_NAME)
        self.assertEqual(model.geoid_directory(), path)

    def test_custom_path(self):
        default_path = pathlib.Path(GeoidModel.default_geoid_path())
        with tempfile.TemporaryDirectory() as dirname:
            geoid_path = pathlib.Path(dirname) / default_path.name
            geoid_path.mkdir()
            for filename in default_path.glob(f"{self.MODEL_NAME}*"):
                shutil.copy(filename, geoid_path)

            try:
                model = GeoidModel(self.MODEL_NAME, geoid_path)
                # , threadsafe=True)
                self.assertEqual(model.geoid_name(), self.MODEL_NAME)
                self.assertEqual(model.geoid_directory(), str(geoid_path))
            finally:
                del model  # release resources

    def test_custom_path_from_env01(self):
        default_path = pathlib.Path(GeoidModel.default_geoid_path())
        with tempfile.TemporaryDirectory() as dirname:
            geoid_path = pathlib.Path(dirname) / default_path.name
            geoid_path.mkdir()
            for filename in default_path.glob(f"{self.MODEL_NAME}*"):
                shutil.copy(filename, geoid_path)

            old_env = os.environ.get("GEOGRAPHICLIB_DATA")
            os.environ["GEOGRAPHICLIB_DATA"] = dirname
            try:
                model = GeoidModel(self.MODEL_NAME)  # , threadsafe=True)
                self.assertEqual(model.geoid_name(), self.MODEL_NAME)
                self.assertEqual(
                    pathlib.Path(model.geoid_directory()), geoid_path
                )
            finally:
                del model  # release resources
                if old_env is None:
                    del os.environ["GEOGRAPHICLIB_DATA"]
                else:
                    os.environ["GEOGRAPHICLIB_DATA"] = old_env

    def test_custom_path_from_env02(self):
        default_path = pathlib.Path(GeoidModel.default_geoid_path())
        with tempfile.TemporaryDirectory() as dirname:
            geoid_path = pathlib.Path(dirname) / default_path.name
            geoid_path.mkdir()
            for filename in default_path.glob(f"{self.MODEL_NAME}*"):
                shutil.copy(filename, geoid_path)

            old_env = os.environ.get("GEOGRAPHICLIB_GEOID_PATH")
            os.environ["GEOGRAPHICLIB_GEOID_PATH"] = str(geoid_path)
            try:
                model = GeoidModel(self.MODEL_NAME)  # , threadsafe=True)
                self.assertEqual(model.geoid_name(), self.MODEL_NAME)
                self.assertEqual(model.geoid_directory(), str(geoid_path))
            finally:
                del model  # release resources
                if old_env is None:
                    del os.environ["GEOGRAPHICLIB_GEOID_PATH"]
                else:
                    os.environ["GEOGRAPHICLIB_GEOID_PATH"] = old_env

    def test_cubic_true(self):
        path = GeoidModel.default_geoid_path()
        model = GeoidModel(self.MODEL_NAME, path, cubic=True)
        self.assertEqual(model.geoid_name(), self.MODEL_NAME)
        self.assertEqual(model.geoid_directory(), path)
        self.assertEqual(model.interpolation(), "cubic")

    def test_cubic_false(self):
        path = GeoidModel.default_geoid_path()
        model = GeoidModel(self.MODEL_NAME, path, cubic=False)
        self.assertEqual(model.geoid_name(), self.MODEL_NAME)
        self.assertEqual(model.geoid_directory(), path)
        self.assertEqual(model.interpolation(), "bilinear")

    def test_threadsafe_true(self):
        path = GeoidModel.default_geoid_path()
        model = GeoidModel(self.MODEL_NAME, path, threadsafe=True)
        self.assertEqual(model.geoid_name(), self.MODEL_NAME)
        self.assertEqual(model.geoid_directory(), path)
        self.assertTrue(model.threadsafe())

    def test_threadsafe_false(self):
        path = GeoidModel.default_geoid_path()
        model = GeoidModel(self.MODEL_NAME, path, threadsafe=False)
        self.assertEqual(model.geoid_name(), self.MODEL_NAME)
        self.assertEqual(model.geoid_directory(), path)
        self.assertFalse(model.threadsafe())

    def test_positional_true_false(self):
        path = GeoidModel.default_geoid_path()
        model = GeoidModel(self.MODEL_NAME, path, cubic=True, threadsafe=False)
        self.assertEqual(model.geoid_name(), self.MODEL_NAME)
        self.assertEqual(model.geoid_directory(), path)
        self.assertEqual(model.interpolation(), "cubic")
        self.assertFalse(model.threadsafe())

    def test_positional_false_true(self):
        path = GeoidModel.default_geoid_path()
        model = GeoidModel(self.MODEL_NAME, path, cubic=False, threadsafe=True)
        self.assertEqual(model.geoid_name(), self.MODEL_NAME)
        self.assertEqual(model.geoid_directory(), path)
        self.assertEqual(model.interpolation(), "bilinear")
        self.assertTrue(model.threadsafe())


class InfoMethodsTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.name = GeoidModel.default_geoid_name()
        self.datapath = GeoidModel.default_geoid_path()
        self.model = GeoidModel(self.name, self.datapath)

    def test_description(self):
        description = self.model.description()
        self.assertIsInstance(description, str)
        self.assertNotEqual(description, "NONE")

    def test_datetime(self):
        datestr = self.model.datetime()
        self.assertIsInstance(datestr, str)
        self.assertNotEqual(datestr, "UNKNOWN")
        # date = datetime.datetime.strptime(datestr, '%Y-%m-%d')
        date = datetime.datetime.strptime(datestr, "%Y-%m-%d %H:%M:%S")
        self.assertLess(date, datetime.datetime.now())

    def test_geoid_file(self):
        filename = self.model.geoid_file()
        self.assertIn(self.name, filename)
        self.assertIn(self.datapath, filename)

    def test_geoid_name(self):
        name = self.model.geoid_name()
        self.assertEqual(name, self.name)

    def test_geoid_directory(self):
        path = self.model.geoid_directory()
        self.assertEqual(path, self.datapath)

    def test_interpolation(self):
        self.assertIn(self.model.interpolation(), ["cubic", "bilinear"])

    def test_threadsafe(self):
        self.assertIsInstance(self.model.threadsafe(), bool)

    def test_max_error(self):
        self.assertIsInstance(self.model.max_error(), float)
        self.assertGreater(self.model.max_error(), 0)

    def test_rms_error(self):
        self.assertIsInstance(self.model.rms_error(), float)
        self.assertGreater(self.model.rms_error(), 0)

    def test_max_vs_rms_error(self):
        self.assertLessEqual(self.model.rms_error(), self.model.max_error())

    def test_offset(self):
        self.assertIsInstance(self.model.offset(), float)

    def test_scale(self):
        self.assertIsInstance(self.model.scale(), float)
        self.assertGreater(self.model.scale(), 0)

    def test_equatorial_radius(self):
        self.assertIsInstance(self.model.equatorial_radius(), float)
        self.assertGreater(self.model.equatorial_radius(), 5e6)

    def test_flattening(self):
        self.assertIsInstance(self.model.flattening(), float)
        self.assertGreater(self.model.flattening(), 0)
        self.assertLess(self.model.flattening(), 1)


class CacheTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.name = GeoidModel.default_geoid_name()
        self.model = GeoidModel(self.name)

    def test_cache_clear(self):
        self.model.cache_clear()
        self.assertFalse(self.model.cache())
        self.model.cache_clear()
        self.assertFalse(self.model.cache())

    def test_cache_all(self):
        self.model.cache_clear()
        self.assertFalse(self.model.cache())
        self.model.cache_all()
        self.assertTrue(self.model.cache())
        self.assertEqual(self.model.cache_south(), -90)
        self.assertEqual(self.model.cache_west(), 0)
        self.assertEqual(self.model.cache_north(), +90)
        self.assertEqual(self.model.cache_east(), 360)
        self.model.cache_clear()

    def test_cache_area(self):
        south = -10
        west = -10
        north = 10
        east = 10

        self.model.cache_clear()
        self.assertFalse(self.model.cache())
        self.model.cache_area(south, west, north, east)
        self.assertTrue(self.model.cache())
        self.assertAlmostEqual(self.model.cache_south(), south, delta=0.1)
        self.assertAlmostEqual(self.model.cache_west(), west, delta=0.1)
        self.assertAlmostEqual(self.model.cache_north(), north, delta=0.1)
        self.assertAlmostEqual(self.model.cache_east(), east, delta=0.1)
        self.model.cache_clear()

    def test_cache(self):
        self.model.cache_clear()
        self.assertIsInstance(self.model.cache(), bool)
        self.assertFalse(self.model.cache())

    def test_cache_south(self):
        self.model.cache_all()
        self.assertIsInstance(self.model.cache_south(), float)
        self.assertAlmostEqual(self.model.cache_south(), -90)
        self.model.cache_clear()

    def test_cache_west(self):
        self.model.cache_all()
        self.assertIsInstance(self.model.cache_west(), float)
        self.assertAlmostEqual(self.model.cache_west(), 0)
        self.model.cache_clear()

    def test_cache_north(self):
        self.model.cache_all()
        self.assertIsInstance(self.model.cache_north(), float)
        self.assertAlmostEqual(self.model.cache_north(), +90)
        self.model.cache_clear()

    def test_cache_east(self):
        self.model.cache_all()
        self.assertIsInstance(self.model.cache_east(), float)
        self.assertAlmostEqual(self.model.cache_east(), 360)
        self.model.cache_clear()


class ComputationTestCase(unittest.TestCase):
    LAT = +dms_to_dec(16, 46, 33)  # 16:46:33N
    LON = -dms_to_dec(3, 0, 34)  # 03:00:34W
    HEIGHT = +28.7068  # [m]
    GHEIGHT = +300  # [m]
    EHEIGHT = HEIGHT + GHEIGHT  # [m]

    # $ echo 16:46:33N 3:00:34W | GeoidEval
    # 28.7068
    # $ echo 16.775833333333335 -3.0094444444444446 | GeoidEval
    # 28.7068

    def setUp(self) -> None:
        self.name = GeoidModel.default_geoid_name()
        self.model = GeoidModel(self.name)

    def test_compute_scalar(self):
        h = self.model(self.LAT, self.LON)
        npt.assert_allclose(h, self.HEIGHT)

    def test_geoid_to_ellipsoid_scalar(self):
        dir_ = EHeightConvDir.GEOIDTOELLIPSOID
        h = self.model.convert_height(self.LAT, self.LON, self.GHEIGHT, dir_)
        npt.assert_allclose(h, self.EHEIGHT)

    def test_ellipsoid_to_geoid_scalar(self):
        dir_ = EHeightConvDir.ELLIPSOIDTOGEOID
        h = self.model.convert_height(self.LAT, self.LON, self.EHEIGHT, dir_)
        npt.assert_allclose(h, self.GHEIGHT)

    def test_convert_height_dir_none(self):
        dir_ = EHeightConvDir.NONE
        h = self.model.convert_height(self.LON, self.LAT, self.EHEIGHT, dir_)
        npt.assert_allclose(h, self.EHEIGHT)


class VectorComputationTestCase(unittest.TestCase):
    LAT = np.asarray(
        [
            [+dms_to_dec(16, 46, 33), -dms_to_dec(16, 46, 43)],
            [-dms_to_dec(16, 56, 33), +dms_to_dec(16, 56, 43)],
        ]
    )
    LON = np.asarray(
        [
            [-dms_to_dec(3, 0, 34), +dms_to_dec(3, 0, 44)],
            [+dms_to_dec(3, 10, 34), -dms_to_dec(3, 10, 44)],
        ]
    )
    HEIGHT = np.asarray(
        [
            [+28.7068, +14.8866],
            [+15.0314, +28.6599],
        ]
    )
    GHEIGHT = np.asarray(
        [
            [+300, +400000],
            [+400000, +300],
        ]
    )
    EHEIGHT = HEIGHT + GHEIGHT
    RTOL = 2.5e-6

    def setUp(self) -> None:
        self.name = GeoidModel.default_geoid_name()
        self.model = GeoidModel(self.name)

    def test_compute_vector(self):
        h = self.model(self.LAT.flatten(), self.LON.flatten())
        npt.assert_allclose(h, self.HEIGHT.flatten(), rtol=self.RTOL)

    def test_compute_matrix(self):
        h = self.model(self.LAT, self.LON)
        npt.assert_allclose(h, self.HEIGHT, rtol=self.RTOL)

    def test_geoid_to_ellipsoid_vector(self):
        dir_ = EHeightConvDir.GEOIDTOELLIPSOID
        h = self.model.convert_height(
            self.LAT.flatten(),
            self.LON.flatten(),
            self.GHEIGHT.flatten(),
            dir_,
        )
        npt.assert_allclose(h, self.EHEIGHT.flatten(), rtol=self.RTOL)

    def test_geoid_to_ellipsoid_matrix(self):
        dir_ = EHeightConvDir.GEOIDTOELLIPSOID
        h = self.model.convert_height(self.LAT, self.LON, self.GHEIGHT, dir_)
        npt.assert_allclose(h, self.EHEIGHT, rtol=self.RTOL)

    def test_ellipsoid_to_geoid_vector(self):
        dir_ = EHeightConvDir.ELLIPSOIDTOGEOID
        h = self.model.convert_height(
            self.LAT.flatten(),
            self.LON.flatten(),
            self.EHEIGHT.flatten(),
            dir_,
        )
        npt.assert_allclose(h, self.GHEIGHT.flatten(), rtol=self.RTOL)

    def test_ellipsoid_to_geoid_matrix(self):
        dir_ = EHeightConvDir.ELLIPSOIDTOGEOID
        h = self.model.convert_height(self.LAT, self.LON, self.EHEIGHT, dir_)
        npt.assert_allclose(h, self.GHEIGHT, rtol=self.RTOL)

    def test_convert_height_dir_none_vector(self):
        dir_ = EHeightConvDir.NONE
        h = self.model.convert_height(
            self.LON.flatten(),
            self.LAT.flatten(),
            self.EHEIGHT.flatten(),
            dir_,
        )
        npt.assert_allclose(h, self.EHEIGHT.flatten())

    def test_convert_height_dir_none_matrix(self):
        dir_ = EHeightConvDir.NONE
        h = self.model.convert_height(self.LON, self.LAT, self.EHEIGHT, dir_)
        npt.assert_allclose(h, self.EHEIGHT)


class ConstHeightVectorComputationTestCase(unittest.TestCase):
    LAT = np.asarray(
        [
            [+dms_to_dec(16, 46, 33), +dms_to_dec(16, 56, 43)],
            [+dms_to_dec(16, 46, 33), +dms_to_dec(16, 56, 43)],
        ]
    )
    LON = np.asarray(
        [
            [-dms_to_dec(3, 0, 34), -dms_to_dec(3, 10, 44)],
            [-dms_to_dec(3, 0, 34), -dms_to_dec(3, 10, 44)],
        ]
    )
    HEIGHT = np.asarray(
        [
            [+28.7068, +28.6599],
            [+28.7068, +28.6599],
        ]
    )
    KHEIGHT = +300
    RTOL = 2.5e-6

    def setUp(self) -> None:
        self.name = GeoidModel.default_geoid_name()
        self.model = GeoidModel(self.name)

    def test_geoid_to_ellipsoid_vector(self):
        dir_ = EHeightConvDir.GEOIDTOELLIPSOID
        h = self.model.convert_height(
            self.LAT.flatten(), self.LON.flatten(), self.KHEIGHT, dir_
        )
        eheight = self.HEIGHT + self.KHEIGHT
        npt.assert_allclose(h, eheight.flatten(), rtol=self.RTOL)

    def test_geoid_to_ellipsoid_matrix(self):
        dir_ = EHeightConvDir.GEOIDTOELLIPSOID
        h = self.model.convert_height(self.LAT, self.LON, self.KHEIGHT, dir_)
        eheight = self.HEIGHT + self.KHEIGHT
        npt.assert_allclose(h, eheight, rtol=self.RTOL)

    def test_ellipsoid_to_geoid_vector(self):
        dir_ = EHeightConvDir.ELLIPSOIDTOGEOID
        h = self.model.convert_height(
            self.LAT.flatten(), self.LON.flatten(), self.KHEIGHT, dir_
        )
        gheight = self.KHEIGHT - self.HEIGHT
        npt.assert_allclose(h, gheight.flatten(), rtol=self.RTOL)

    def test_ellipsoid_to_geoid_matrix(self):
        dir_ = EHeightConvDir.ELLIPSOIDTOGEOID
        h = self.model.convert_height(self.LAT, self.LON, self.KHEIGHT, dir_)
        gheight = self.KHEIGHT - self.HEIGHT
        npt.assert_allclose(h, gheight, rtol=self.RTOL)

    def test_convert_height_dir_none_vector(self):
        dir_ = EHeightConvDir.NONE
        h = self.model.convert_height(
            self.LON.flatten(), self.LAT.flatten(), self.KHEIGHT, dir_
        )
        npt.assert_allclose(h, np.full_like(h, self.KHEIGHT))

    def test_convert_height_dir_none_matrix(self):
        dir_ = EHeightConvDir.NONE
        h = self.model.convert_height(self.LON, self.LAT, self.KHEIGHT, dir_)
        npt.assert_allclose(h, np.full_like(h, self.KHEIGHT))



"""Tools for geographic models data download and installation."""

import os
import enum
import shutil
import logging
import pathlib
import tempfile
import contextlib
from typing import Union
from urllib.parse import urlsplit
from urllib.request import urlretrieve
from collections.abc import Callable, Iterable

from ._typing import PathType

__all__ = [
    "EModelGroup",
    "EModelType",
    "EGeoidModel",
    "EGravityModel",
    "EMagneticModel",
    "EArchiveType",
    "get_default_data_path",
    "get_model_url",
    "install",
]


class EModelGroup(enum.Enum):
    """Model groups."""

    ALL = "all"
    MINIMAL = "minimal"
    RECOMMENDED = "recommended"


class EModelType(enum.Enum):
    """Enumerate geographic model types."""

    GEOID = "geoids"
    GRAVITY = "gravity"
    MAGNETIC = "magnetic"


class EGeoidModel(enum.Enum):
    """Enumerate geoid models."""

    EGM84_30 = "egm84-30"
    EGM84_15 = "egm84-15"
    EGM96_15 = "egm96-15"
    EGM96_5 = "egm96-5"
    EGM2008_5 = "egm2008-5"
    EGM2008_2_5 = "egm2008-2_5"
    EGM2008_1 = "egm2008-1"

    @staticmethod
    def get_model_type() -> EModelType:
        """Return the model type corresponding to the enumeration."""
        return EModelType.GEOID


class EGravityModel(enum.Enum):
    """Enumerate gravity models."""

    EGM84 = "egm84"
    EGM96 = "egm96"
    EGM2008 = "egm2008"
    GRS80 = "grs80"
    WGS84 = "wgs84"

    @staticmethod
    def get_model_type() -> EModelType:
        """Return the model type corresponding to the enumeration."""
        return EModelType.GRAVITY


class EMagneticModel(enum.Enum):
    """Enumerate magnetic field models."""

    WMM2010 = "wmm2010"
    WMM2015 = "wmm2015"
    WMM2015V2 = "wmm2015v2"
    WMM2020 = "wmm2020"
    WMM2025 = "wmm2025"
    WMMHR2025 = "wmmhr2025"
    IGRF11 = "igrf11"
    IGRF12 = "igrf12"
    IGRF13 = "igrf13"
    IGRF14 = "igrf14"
    EMM2010 = "emm2010"
    EMM2015 = "emm2015"
    EMM2017 = "emm2017"

    @staticmethod
    def get_model_type() -> EModelType:
        """Return the model type corresponding to the enumeration."""
        return EModelType.MAGNETIC


class EArchiveType(enum.Enum):
    """Enumerate the archive type."""

    ZIP = ".zip"
    BZ2 = ".tar.bz2"


GenericModelType = Union[EGeoidModel, EGravityModel, EMagneticModel]
ReportHookType = Callable[[int, int, int], None]


def get_default_data_path() -> str:
    """Return the default data path.

    The `GEOGRAPHICLIB_DATA` environment variable is used if available
    to locate the location where the geographic model data are installed.

    If `GEOGRAPHICLIB_DATA` is not set, then then it is returned the
    path configured at build time.
    """
    path = os.environ.get("GEOGRAPHICLIB_DATA")
    if path is None:
        from . import MagneticFieldModel

        path = os.path.dirname(MagneticFieldModel.default_magnetic_path())
    return path


_BASE_URL = "https://downloads.sourceforge.net/project/geographiclib/"
_URL_PATH_TEMPLATE = "{basepath}/{modeltype}-distrib/{filename}{ext}"
_URL_QUERY = "use_mirror=autoselect"
_URL_FRAGMENT = ""


def get_base_url() -> str:
    """Return the base URL for data download."""
    return _BASE_URL


def get_model_url(
    model: GenericModelType,
    base_url: str | None = None,
    archive_type: EArchiveType = EArchiveType.BZ2,
) -> str:
    """Return the download URL for the specified geographic model.

    :param model:
        the enumeration corresponding to the desired geographic model.
        It can be one of the enumerates defined in :class:`EGeoidModel`,
        :class:`EGravityModel` or :class:`EMagneticModel`.
    :param str base_url:
        (optional) base URL for data download.
        The full URL is build from this function starting from base_url
        and model information.
    :param EArchiveType archive_type:
        specifies the archive type that should be downloaded.
        Default: `data:`EArchiveType.BZ2`.
    """
    if not base_url:
        base_url = get_base_url()
        query = _URL_QUERY
        fragment = _URL_FRAGMENT
    else:
        query = ""
        fragment = ""

    url = urlsplit(base_url)
    urlpath = _URL_PATH_TEMPLATE.format(
        basepath=url.path.rstrip("/"),
        modeltype=model.get_model_type().value,
        filename=model.value,
        ext=archive_type.value,
    )
    url = url._replace(path=urlpath, query=query, fragment=fragment)
    return url.geturl()


InstallableModelType = Union[EModelGroup, EModelType, GenericModelType]


_MODELTYPE_MAP: dict[EModelType, type[GenericModelType]] = {
    EModelType.GEOID: EGeoidModel,
    EModelType.GRAVITY: EGravityModel,
    EModelType.MAGNETIC: EMagneticModel,
}


def _get_url_map_from_group(
    model: EModelGroup,
    base_url: str | None = None,
    archive_type: EArchiveType = EArchiveType.BZ2,
) -> dict[GenericModelType, str]:
    urls: dict[GenericModelType, str] = {}
    match model:
        case EModelGroup.ALL:
            for modeltype in EModelType:
                urls.update(_get_url_map(modeltype, base_url, archive_type))
        case EModelGroup.MINIMAL:
            from . import GeoidModel, GravityModel, MagneticFieldModel

            models: list[GenericModelType] = [
                EGeoidModel(GeoidModel.default_geoid_name()),
                EGravityModel(GravityModel.default_gravity_name()),
                EMagneticModel(MagneticFieldModel.default_magnetic_name()),
            ]
            for model_ in models:
                urls[model_] = get_model_url(model_, base_url, archive_type)
        case EModelGroup.RECOMMENDED:
            urls.update(_get_url_map(EModelGroup.MINIMAL))
            extra_models: list[GenericModelType] = [
                EGeoidModel.EGM96_5,
                EGravityModel.EGM96,
                EMagneticModel.IGRF12,
                EMagneticModel.WMM2015,
            ]
            for model_ in extra_models:
                urls[model_] = get_model_url(model_, base_url, archive_type)
        case _:
            raise ValueError(f"unexpected value: {model!r}")

    return urls


def _get_url_map(
    model: InstallableModelType,
    base_url: str | None = None,
    archive_type: EArchiveType = EArchiveType.BZ2,
) -> dict[GenericModelType, str]:
    urls: dict[GenericModelType, str]
    match model:
        case EModelGroup() as group:
            urls = _get_url_map_from_group(group)
        case EModelType() as modeltype:
            models: Iterable[GenericModelType] = _MODELTYPE_MAP[modeltype]
            urls = {
                model_: get_model_url(model_, base_url, archive_type)
                for model_ in models
            }
        case EGeoidModel() | EGravityModel() | EMagneticModel() as mod:
            urls = {mod: get_model_url(mod, base_url, archive_type)}
        case _:
            raise ValueError(f"unexpected model: {model!r}")

    return urls


have_tqdm: bool
try:
    import tqdm

    have_tqdm = True

    class TqdmReportHook(tqdm.tqdm):
        """Tqdm based report hook for urllib.request.urlretrieve."""

        def __init__(self, **kwargs):
            if "iterable" in kwargs:
                raise TypeError(
                    "{!r} argument is not allowed by TqdmReportHook."
                )

            # set defaults
            kargs = {
                "unit": "B",
                "unit_scale": True,
                "unit_divisor": 1024,
                "miniters": 1,
            }
            kargs.update(kwargs)

            super().__init__(**kargs)  # type: ignore[call-overload]

        def __call__(
            self,
            count: int = 1,
            block_size: int = 1,
            total_size: int | None = None,
        ):
            if total_size not in (None, -1):
                self.total = total_size

            # will also set: self.n = cont * block_size
            self.update(count * block_size - self.n)

        def __bool__(self):
            # needed by urllib.request.urlretrieve
            return True

except ImportError:
    have_tqdm = False


def _get_report_hook(
    progress: bool | ReportHookType, description: str
) -> ReportHookType | None:
    if progress is True and have_tqdm:
        try:
            base_width = 50
            ncols = os.get_terminal_size().columns
            ncols = ncols - base_width if ncols >= base_width else 0
        except OSError:
            ncols = 0
        return TqdmReportHook(desc=description[-ncols:], leave=False)
    elif callable(progress):
        return progress
    else:
        return None


def download(
    url: str,
    path: PathType = ".",
    progress: bool | ReportHookType = True,
    force: bool = False,
) -> str:
    """Download the specified URL.

    :param str url:
        the url of the file to be downloaded (simple file names are also OK)
    :param str path:
        (optional) target download location (filename or directory).
        Default: ".".
    :param progress:
        boolean to enable/disable the display of progress information
        (default: True).
        Please note that default progress reporting uses the 'tqdm'
        library, if it is not available progress display is switched off.
        It `progress` is set to a callable it is used ar reporthook to show
        progress information.
        Please refer to the :func:`urllib.request.urlretrieve` for
        details about the reporthook function.
    :param bool force:
        overwrite existing files (default: False)
    :returns:
        pathname of the downloaded file
    """
    urlobj = urlsplit(url)
    if not urlobj.scheme:
        urlobj = urlobj._replace(scheme="file")

    path = pathlib.Path(path)
    if path.is_dir():
        path /= pathlib.Path(urlobj.path).name

    if path.exists() and not force:
        raise RuntimeError(f'download target path already exists: "{path}"')

    desc = urlsplit(url)._replace(query="", fragment="").geturl()
    report_hook = _get_report_hook(progress=progress, description=desc)

    if isinstance(report_hook, contextlib.AbstractContextManager):
        with report_hook:
            outpath, _ = urlretrieve(urlobj.geturl(), path, report_hook)
    else:
        outpath, _ = urlretrieve(urlobj.geturl(), path, report_hook)

    return outpath


def install(
    model: InstallableModelType = EModelGroup.MINIMAL,
    datadir: PathType | None = None,
    base_url: str | None = None,
    archive_type: EArchiveType = EArchiveType.BZ2,
    progress: bool = True,
):
    """Install the specified geographic model data.

    :param model:
        the enumeration corresponding to the desired geographic model.
        It can be one of the enumerates defined in :class:`EGeoidModel`,
        :class:`EGravityModel`, :class:`EMagneticModel`, or one of the
        enumerates defined in :class:`EModelType` to indicate that all
        geographic models of the specified type shall be installed,
        or one of the enumerates defined in :class:`EModelGroup` to
        indicate that a specific group of model data shall be installed:
        :data:`EModelGroup.ALL` (all available models of any kind),
        :data:`EModelGroup.MINIMAL` (only the default model for each type)
        or :data:`EModelGroup.RECOMMENDED`.
    :param PathType datadir:
        (optional) specify the target location where geographic model
        data shall be installed. If not specified that the path returned
        by :func:`get_default_data_path` is assumed.
    :param str base_url:
        (optional) base URL for data download.
        The full URL is built from this function starting from base_url
        and model information.
    :param EArchiveType archive_type:
        specifies the archive type that should be downloaded.
        Default: :data:`EArchiveType.BZ2`.
    :param bool progress:
        enable/disable progress information display (default: True)
    """
    urls = _get_url_map(model, base_url, archive_type)

    if not datadir:
        datadir = get_default_data_path()
        if datadir is None:
            raise RuntimeError("no default data location found")

        datadir = pathlib.Path(datadir)
        if not datadir.is_dir():
            raise NotADirectoryError(f'"{datadir}" is not a directory')

    datadir = pathlib.Path(datadir)
    datadir.mkdir(parents=True, exist_ok=True)

    with tempfile.TemporaryDirectory() as tempdir:
        urliterator: Iterable[tuple[GenericModelType, str]]
        if progress and have_tqdm and len(urls) > 1:
            urliterator = tqdm.tqdm(urls.items(), unit="file", desc="download")
        else:
            urliterator = urls.items()

        for model, url in urliterator:
            target = datadir / model.get_model_type().value / model.value
            matches = list(target.parent.glob(f"{target.name}*"))
            if matches:
                logging.debug('"%s" already exists: skip download', target)
                continue
            filename = download(url, tempdir, progress=progress)
            # @CMPATIBILITY: shutil.unpack_archive accepts pathlib.Path
            #                parameters since Python 3.7
            shutil.unpack_archive(str(filename), extract_dir=str(datadir))



"""World Magnetic Model Format tools.

Classes and function for reading and writing files in World Magnetic
Model Format and in IGRF text format.

See: https://geographiclib.sourceforge.io/C++/doc/magnetic.html#magneticformat
and https://www.ngdc.noaa.gov/IAGA/vmod/igrf.html.
"""

from __future__ import annotations

import struct
import typing
import fnmatch
import pathlib
import datetime
import warnings
from collections import OrderedDict, namedtuple
from dataclasses import dataclass
from urllib.parse import urlsplit
from urllib.request import urlopen

import numpy as np

from ._typing import PathType

__all__ = ["MetaData", "SphCoeffSet", "WmmData", "import_igrf_txt"]


@dataclass
class MetaData:
    """World Magnetic Model Format metadata."""

    Name: str = "N/A"
    Description: str = "International Geomagnetic Reference Field"
    URL: str = "http://ngdc.noaa.gov/IAGA/vmod/igrf.html"
    Publisher: str = "International Association of Geomagnetism and Aeronomy"
    ReleaseDate: str = "N/A"
    DataCutOff: str = "N/A"
    ConversionDate: str = "N/A"
    DataVersion: int = 1
    Radius: int = 6371200
    NumModels: int = 0
    Epoch: int = 1900
    DeltaEpoch: int = 5
    MinTime: int = 0
    MaxTime: int = 0
    MinHeight: int = -1000
    MaxHeight: int = 600000
    NumConstants: int = 0
    Normalization: int = 1
    ID: str = "N/A"
    Type = "linear"
    ByteOrder = "little"
    N: int = 0
    M: int = 0
    FORMAT_VERSION: int = 1

    _DATEFMT = "%Y-%m-%d"

    def load(self, filename: PathType) -> None:
        """Load a metadata file in WMM format and return a MetaData object."""
        type_map = typing.get_type_hints(self)
        with open(filename) as fd:
            line = next(fd)
            format_ = line.rstrip()
            if format_ not in ("WMMF-1", "WMMF-2"):
                raise RuntimeError(f"invalid format: {format_!r}")
            else:
                self.FORMAT_VERSION = int(format_[-1])

            for line in fd:
                line = line.split("#", 1)[0]
                line = line.strip()
                if not line:
                    continue

                name, value = line.split(None, 1)
                if not hasattr(self, name):
                    warnings.warn(f"unexpected field: {name!r}")
                    continue
                if callable(getattr(self, name)):
                    warnings.warn(
                        f"field name conflicts with method name: {name!r}"
                    )
                    continue
                field_type = type_map.get(name, str)
                setattr(self, name, field_type(value))

    def get_years(self) -> list[int]:
        """Return the list of years associated to the models."""
        years = range(
            self.Epoch,
            self.Epoch + self.DeltaEpoch * self.NumModels,
            self.DeltaEpoch,
        )
        return list(years)

    def _generation_str(self) -> str:
        if self.Name and self.Name != "N/A":
            generation = str(int(self.Name[-2:]) - 1)
            if generation == "1":
                generation = f"{generation}st Generation"
            elif generation == "2":
                generation = f"{generation}nd Generation"
            elif generation == "3":
                generation = f"{generation}rd Generation"
            else:
                assert "0" <= generation[-1] <= "9"
                generation = f"{generation}th Generation"
        else:
            generation = ""
        return generation

    def get_id(self) -> str:
        """Return the model ID (eventually inferred by the model name)."""
        if not self.ID or self.ID == "N/A":
            if not self.Name.startswith("igrf") or len(self.Name) != 6:
                id_ = "N/A"
            else:
                id_ = self.Name.upper() + "-A"
        else:
            id_ = self.ID
        return id_

    def __str__(self) -> str:
        """Return the string representation of the MetaDAta object."""
        upper_name = self.Name.upper()

        description = self.Description
        if description == "International Geomagnetic Reference Field":
            gen = self._generation_str()
            description = f"{description} {gen}"

        id_ = self.get_id()

        return f"""\
WMMF-{self.FORMAT_VERSION}
# A World Magnetic Model (Format {self.FORMAT_VERSION}) file.  \
For documentation on the
# format of this file see
# http://geographiclib.sf.net/html/magnetic.html#magneticformat
Name            {self.Name}
Description     {description}
URL             {self.URL}
Publisher       {self.Publisher}
ReleaseDate     {self.ReleaseDate}
DataCutOff      {self.DataCutOff}
ConversionDate  {self.ConversionDate}
DataVersion     {self.DataVersion}
Radius          {self.Radius}
NumModels       {self.NumModels}
Epoch           {self.Epoch}
DeltaEpoch      {self.DeltaEpoch}
MinTime         {self.MinTime}
MaxTime         {self.MaxTime}
MinHeight       {self.MinHeight}
MaxHeight       {self.MaxHeight}

# The coefficients are stored in a file obtained by appending ".cof" to
# the name of this file.  The coefficients were obtained from {upper_name}.COF
# in the geomag70 distribution.
ID              {id_}
"""

    def save(self, filename: PathType) -> None:
        """Save metadata in WMM format."""
        with open(filename, "w") as fd:
            fd.write(str(self))


SphCoeffSet = namedtuple("SphCoeffSet", ["C", "S"])
SphCoeffsType = OrderedDict[str, SphCoeffSet]


class WmmData:
    """Magnetic field data."""

    @classmethod
    def from_metadata_and_coeffs(
        cls, medadata: MetaData, coeffs: SphCoeffsType
    ) -> WmmData:
        """Instantiate a WmmData object from metadata and coefficients."""
        wmmdata = cls()
        wmmdata.metadata = medadata
        wmmdata.coeffs = coeffs
        wmmdata._check()
        return wmmdata

    def __init__(self, filename: PathType | None = None) -> None:
        """Initialize a WmmData instance."""
        self.metadata = MetaData()
        self.coeffs: SphCoeffsType = OrderedDict()

        if filename:
            filename = pathlib.Path(filename)
            self.load(filename)

    def _check(self) -> None:
        years = list(self.coeffs.keys())
        rate = years.pop()
        if rate != "rate":
            raise RuntimeError("rate coefficients not found")
        if years != [str(year) for year in self.metadata.get_years()]:
            raise RuntimeError("coefficient years does not match metadata")

    @staticmethod
    def _load_sph_coeff_set(fd: typing.IO) -> SphCoeffSet:
        data = fd.read(2 * 4)
        n, m = struct.unpack("<ii", data)

        nc = (m + 1) * (2 * n - m + 2) // 2
        data = np.fromfile(fd, dtype=np.float64, count=nc)
        cosine_coef = np.zeros((n + 1, m + 1))
        cosine_coef.T[np.triu_indices(m + 1, 0, n + 1)] = data

        nc = m * (2 * n - m + 1) // 2
        data = np.fromfile(fd, dtype=np.float64, count=nc)
        sine_coef = np.zeros((n + 1, m + 1))
        sine_coef.T[1:, 1:][np.triu_indices(m, 0, n)] = data

        return SphCoeffSet(cosine_coef, sine_coef)

    def _load_sph_coeffs(self, filename: PathType) -> SphCoeffsType:
        """Load spherical harmonics coeffs from a binary file in WMM format."""
        filename = pathlib.Path(filename)
        years = self.metadata.get_years()

        with open(filename, "rb") as fd:
            id_ = fd.read(8).decode("utf-8")
            if id_ != self.metadata.ID:
                raise RuntimeError(
                    f"data ID ({id_}) does not match "
                    f"metadata ID ({self.metadata.ID})"
                )
            data = OrderedDict()
            for year in years:
                data[str(year)] = self._load_sph_coeff_set(fd)
            data["rate"] = self._load_sph_coeff_set(fd)

        return data

    def load(self, filename: PathType) -> None:
        """Load metadata and spherical harmonics coefficients."""
        filename = pathlib.Path(filename)
        # self.metadata = MetaData()  # @TODO: check
        self.metadata.load(filename)
        bin_filename = filename.with_suffix(filename.suffix + ".cof")
        self.coeffs = self._load_sph_coeffs(bin_filename)
        self._check()

    @staticmethod
    def _save_sph_coeff_set(fd: typing.IO, coeffs: SphCoeffSet) -> None:
        if coeffs.C.ndim != 2:
            raise TypeError("coefficients are not 2d arrays")
        if coeffs.C.shape != coeffs.S.shape:
            raise ValueError(
                f"C and S coefficient do not have the same shape "
                f"(C: {coeffs.C.shape}, S: {coeffs.S.shape})"
            )

        n, m = coeffs.C.shape
        n -= 1
        m -= 1
        if m > n:
            raise TypeError(
                f"invalid shape of coefficient arrays: "
                f"n={n}, m={m}, expected m <= n"
            )

        # compute the effective size (n, m)
        size = np.abs(coeffs.C) + np.abs(coeffs.S)
        n = np.where(np.sum(size, 0) > 0)[0][-1]
        m = np.where(np.sum(size, 1) > 0)[0][-1]
        if m > n:
            m = n

        bytes_ = struct.pack("<ii", n, m)
        fd.write(bytes_)

        data = coeffs.C.T[np.triu_indices(m + 1, 0, n + 1)]
        data = np.ascontiguousarray(data, dtype="<f8")
        fd.write(data.tobytes())

        data = coeffs.S.T[1:, 1:][np.triu_indices(m, 0, n)]
        data = np.ascontiguousarray(data, dtype="<f8")
        fd.write(data.tobytes())

    def _save_sph_coeffs(self, filename: PathType) -> None:
        """Store spherical harmonics coefficients in binary WMM format."""
        id_ = self.metadata.get_id()
        if len(id_) != 8:
            raise ValueError(f"invalid id_ ({id_:r}): expected len(id_) == 8")

        with open(filename, "wb") as fd:
            fd.write(id_.encode("utf-8"))
            for coeffs in self.coeffs.values():
                self._save_sph_coeff_set(fd, coeffs)

    def save(self, outpath: PathType, force: bool = False) -> None:
        """Save data in WMM format (metadata and binary)."""
        outpath = pathlib.Path(outpath)
        if outpath.is_dir():
            filename = outpath / (self.metadata.Name + ".wmm")
        else:
            filename = outpath
        bin_filename = filename.with_suffix(filename.suffix + ".cof")

        if not force:
            if filename.exists():
                raise RuntimeError(f'"{filename}" already exists')
            if bin_filename.exists():
                raise RuntimeError(f'"{bin_filename}" already exists')

        self.metadata.save(filename)
        self._save_sph_coeffs(bin_filename)


def _metadata_from_txt_header(header: str) -> MetaData:
    """Build a MataData object from the header of a coeff file.

    The coeff file is expected to be in text IGRF format.
    """
    parts = header.split()
    if parts[:3] != ["g/h", "n", "m"] or "-" not in parts[-1]:
        raise ValueError(f"invalid header line: {header!r}")

    years = [float(item) for item in parts[3:-1]]

    # check uniform time sampling
    dyears = [b - a for a, b in zip(years[1:], years[:-1])]
    if max(dyears) != min(dyears):
        raise RuntimeError("non uniform time sampling detected")

    metadata = MetaData()
    metadata.ConversionDate = datetime.date.today().strftime(MetaData._DATEFMT)
    metadata.NumModels = len(years)
    metadata.Epoch = int(years[0])
    metadata.DeltaEpoch = int(years[1] - years[0])
    metadata.MinTime = int(years[0])
    metadata.MaxTime = int(years[-1]) + metadata.DeltaEpoch

    return metadata


def import_igrf_txt(path: PathType) -> WmmData:
    """Decode a spherical harmonics coefficient file in IGRF text format.

    :param path:
        the path to a local filename or a remote URL.
    """
    filename = pathlib.Path(path)
    urlobj = urlsplit(filename.as_uri())

    pattern = "igrf[0-9][0-9]coeffs.txt"
    if not fnmatch.fnmatch(filename.name, pattern):
        raise ValueError(
            f'invalid file name ("{filename}"), expected pattern: {pattern!r}'
        )

    if urlobj.scheme in ("", "file"):
        fd = open(filename)
    else:
        fd = urlopen(str(path))

    with fd:
        for line in fd:
            if line.startswith("#"):
                continue
            elif line.startswith("g/h"):
                assert filename.stem.endswith("coeffs")
                metadata = _metadata_from_txt_header(line)
                metadata.Name = filename.stem[:-6]
                break
        else:
            raise RuntimeError(f"header line not found in {filename}")

        years = metadata.get_years()
        dtype = np.dtype(
            [
                ("type", "S1"),
                ("n", "int"),
                ("m", "int"),
            ]
            + [(f"{year}", "float64") for year in years]
            + [("rate", "float64")]
        )
        coeffs = np.loadtxt(fd, dtype=dtype)

    n = np.max(coeffs["n"])
    m = np.max(coeffs["m"])

    g_idx = coeffs["type"] == b"g"
    g = coeffs[g_idx]
    h_idx = coeffs["type"] == b"h"
    h = coeffs[h_idx]

    data = OrderedDict()
    for year in years:
        cosine_coef = np.zeros((n + 1, m + 1))
        cosine_coef[g["n"], g["m"]] = g[str(year)]

        sine_coef = np.zeros((n + 1, m + 1))
        sine_coef[h["n"], h["m"]] = h[str(year)]

        data[str(year)] = SphCoeffSet(cosine_coef, sine_coef)

    cosine_coef = np.zeros((n + 1, m + 1))
    cosine_coef[g["n"], g["m"]] = g["rate"]

    sine_coef = np.zeros((n + 1, m + 1))
    sine_coef[h["n"], h["m"]] = h["rate"]

    data["rate"] = SphCoeffSet(cosine_coef, sine_coef)

    return WmmData.from_metadata_and_coeffs(metadata, data)



# -*- coding: utf-8 -*-
# cython: language_level=3
# distutils: language=c++

cdef extern from "GeographicLib/Config.h":
    const int GEOGRAPHICLIB_VERSION_MAJOR
    const int GEOGRAPHICLIB_VERSION_MINOR
    const int GEOGRAPHICLIB_VERSION_PATCH
    const int GEOGRAPHICLIB_VERSION
    const char[] GEOGRAPHICLIB_VERSION_STRING


cdef extern from "GeographicLib/Constants.hpp":
    int GEOGRAPHICLIB_VERSION_NUM(int, int, int)


cdef extern from "GeographicLib/Math.hpp" namespace "GeographicLib":
    cdef cppclass Math:
        ctypedef double real


ctypedef Math.real real



"""Command Line Interface (CLI) for the geomodels Python package."""

import os
import enum
import glob
import logging
import pathlib
import argparse

from . import __version__
from .data import (
    get_default_data_path,
    get_base_url,
    install,
    EModelGroup,
    EModelType,
    EGeoidModel,
    EGravityModel,
    EMagneticModel,
)
from .wmmf import import_igrf_txt
from .tests import print_versions
from ._typing import PathType

EX_FAILURE = 1
EX_INTERRUPT = 130

PROG = __package__ + "-cli"
LOGFMT = "%(levelname)s: %(message)s"
DEFAULT_LOGLEVEL = "WARNING"


class EInfoMode(enum.Enum):
    """Enumeration describing which king of info shall be provided."""

    INFO = "info"
    DATA = "data"
    ALL = "all"


def _autocomplete(parser: argparse.ArgumentParser) -> None:
    try:
        import argcomplete
    except ImportError:
        pass
    else:
        argcomplete.autocomplete(parser)


def _format_data_info(datadir=None):
    if datadir is None:
        datadir = get_default_data_path()

    lines = [f"data directory: {datadir!r}"]
    for modelenum in (EGeoidModel, EGravityModel, EMagneticModel):
        modeltype = modelenum.get_model_type().value
        modeltype_dir = os.path.join(datadir, modeltype)
        lines.append(f"* model: {modeltype} ({modeltype_dir!r})")
        for item in modelenum:
            pattern = os.path.join(modeltype_dir, item.value + "*")
            installed = "INSTALLED " if glob.glob(pattern) else "NOT INSTALLED"
            lines.append(f"  {item.name:12s} - {installed}")

    return "\n".join(lines)


def info(mode=EInfoMode.ALL, datadir=None):
    """Provide information about the installation and environment.

    Information provided include: the platform, the library versions and
    installed data.
    """
    if mode in (EInfoMode.INFO, EInfoMode.ALL):
        print_versions()
    if mode in (EInfoMode.DATA, EInfoMode.ALL):
        print(_format_data_info(datadir))


def install_data(model, datadir=None, base_url=None, no_progress=False):
    """Download and install the data necessary for models computation.

    GeoModels uses external data to perform geoid, gravity and magnetic
    field computations.

    It is possible to install different subsets of data:

    :minimal:
        only data for the default model of each kind (geoid,
        gravity and magnetic field) are installed,
    :recommended:
        install the `minimal` set of data (see above) plus few
        additional and commonly used data (it is guaranteed that
        the `recommended` subset always includes all data that
        are necessary to run the test suite),
    :all:
        install all available data (about 670MB of disk space
        required),
    :geoids:
        install data for all supported geoids,
    :gravity:
        install data for all supported gravity models,
    :magnetic:
        install data for all supported magnetic field models.

    Additionally the it is possible to install data for a single model.
    """
    if datadir is None:
        datadir = get_default_data_path()
    if base_url is None:
        base_url = get_base_url()

    enums = (
        EModelGroup,
        EModelType,
        EGeoidModel,
        EGravityModel,
        EMagneticModel,
    )
    for enumtype in enums:
        try:
            model = enumtype(model)
        except ValueError:
            pass
        else:
            break
    else:
        raise RuntimeError(f"unexpected model: {model!r}")

    progress = not no_progress
    install(model, datadir, base_url, progress=progress)


def import_igrf(
    path: PathType, outpath: PathType | None = None, force: bool = False
):
    """Import magnetic field data from IGRF text format.

    Import Spherical Harmonics coefficients for the IGRF magnetic field
    model from text file in IGRF standard format.

    See: https://www.ngdc.noaa.gov/IAGA/vmod/igrf.html.
    """
    wmmdata = import_igrf_txt(path)

    if outpath is None:
        outpath = pathlib.Path(get_default_data_path()) / "magnetic"

    wmmdata.save(outpath, force)


def _add_logging_control_args(
    parser: argparse.ArgumentParser, default_loglevel: str = DEFAULT_LOGLEVEL
) -> argparse.ArgumentParser:
    """Add command line options for logging control."""
    loglevels = [logging.getLevelName(level) for level in range(10, 60, 10)]

    parser.add_argument(
        "--loglevel",
        default=default_loglevel,
        choices=loglevels,
        help="logging level (default: %(default)s)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="loglevel",
        action="store_const",
        const="ERROR",
        help=(
            "suppress standard output messages, only errors are printed "
            "to screen (set 'loglevel' to 'ERROR')"
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        action="store_const",
        const="INFO",
        help="print verbose output messages (set 'loglevel' to 'INFO')",
    )
    parser.add_argument(
        # "-d",
        "--debug",
        dest="loglevel",
        action="store_const",
        const="DEBUG",
        help="print debug messages (set 'loglevel' to 'DEBUG')",
    )

    return parser


def _get_synopsis(docstring: str | None) -> str:
    return docstring.splitlines()[0] if docstring is not None else ""


def get_info_parser(subparsers=None) -> argparse.ArgumentParser:
    """Set up the argument parser for the `info` sub-command."""
    name = "info"
    synopsis = _get_synopsis(info.__doc__)
    doc = info.__doc__

    if subparsers is None:
        parser = argparse.ArgumentParser(prog=name, description=doc)
    else:
        parser = subparsers.add_parser(name, description=doc, help=synopsis)

    parser.set_defaults(func=info)

    # command line options
    parser.add_argument(
        "-d",
        "--datadir",
        default=get_default_data_path(),
        help=(
            "specifies where the model data are stored (default: %(default)r)."
        ),
    )
    parser.add_argument(
        "-a",
        "--all",
        dest="mode",
        action="store_const",
        const=EInfoMode.ALL,
        default=EInfoMode.INFO,
        help=(
            "show both versions and platform info and also information "
            "about installed data"
        ),
    )
    parser.add_argument(
        "--data",
        dest="mode",
        action="store_const",
        const=EInfoMode.DATA,
        help="show info about installed data",
    )

    return parser


def get_install_data_parser(subparsers=None) -> argparse.ArgumentParser:
    """Set up the argument parser for the `install-data` sub-command."""
    name = "install-data"
    synopsis = _get_synopsis(install_data.__doc__)
    doc = install_data.__doc__

    if subparsers is None:
        parser = argparse.ArgumentParser(
            prog=name,
            description=doc,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = subparsers.add_parser(
            name,
            description=doc,
            help=synopsis,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.set_defaults(func=install_data)

    # command line options
    parser.add_argument(
        "-b",
        "--base-url",
        default=get_base_url(),
        help="specifies the base URL for the download (default: %(default)r).",
    )
    parser.add_argument(
        "-d",
        "--datadir",
        default=get_default_data_path(),
        help=(
            "specifies where the datasets should be stored "
            "(default: %(default)r)."
        ),
    )
    parser.add_argument(
        "--no-progress",
        action="store_true",
        default=False,
        help="suppress progress bar display",
    )

    # positional arguments
    choices = [model.value for model in EModelGroup]
    choices.extend(model.value for model in EModelType)
    choices.extend(model.value for model in EGeoidModel)
    choices.extend(model.value for model in EGravityModel)
    choices.extend(model.value for model in EMagneticModel)
    parser.add_argument(
        "model", choices=choices, help="model(s) to be installed"
    )

    return parser


def get_import_igrf_parser(subparsers=None) -> argparse.ArgumentParser:
    """Set up the  argument parser for the `import-igrf` sub-command."""
    name = "import-igrf"
    synopsis = _get_synopsis(import_igrf.__doc__)
    doc = import_igrf.__doc__

    if subparsers is None:
        parser = argparse.ArgumentParser(prog=name, description=doc)
    else:
        parser = subparsers.add_parser(name, description=doc, help=synopsis)

    parser.set_defaults(func=import_igrf)

    # command line options
    parser.add_argument(
        "-o",
        "--outpath",
        default=pathlib.Path(get_default_data_path()) / "magnetic",
        help='specifies the output data path (default: "%(default)s").',
    )
    parser.add_argument(
        "--force",
        action="store_true",
        default=False,
        help="overwrite existing files (default: %(default)s).",
    )

    # positional arguments
    parser.add_argument("path", help="path or URL of the IGRF text file")

    return parser


def get_parser() -> argparse.ArgumentParser:
    """Instantiate the command line argument (sub-)parser."""
    parser = argparse.ArgumentParser(prog=PROG, description=__doc__)
    parser.add_argument(
        "--version", action="version", version="%(prog)s v" + __version__
    )

    # Command line options
    _add_logging_control_args(parser)

    # Sub-command management
    subparsers = parser.add_subparsers(title="sub-commands")  # , metavar="")
    get_info_parser(subparsers)
    get_install_data_parser(subparsers)
    get_import_igrf_parser(subparsers)

    _autocomplete(parser)

    return parser


def parse_args(args=None, namespace=None, parser=None):
    """Parse command line arguments."""
    if parser is None:
        parser = get_parser()

    args = parser.parse_args(args, namespace)

    # Common pre-processing of parsed arguments and consistency checks
    # ...

    if getattr(args, "func", None) is None:
        parser.error("no sub-command specified.")

    return args


def _get_kwargs(args):
    kwargs = vars(args).copy()
    kwargs.pop("func", None)
    kwargs.pop("loglevel", None)
    return kwargs


def main(*argv):
    """Implement the main CLI interface."""
    # setup logging
    logging.basicConfig(format=LOGFMT, level=DEFAULT_LOGLEVEL)
    logging.captureWarnings(True)
    log = logging.getLogger(__name__)

    # parse cmd line arguments
    args = parse_args(argv if argv else None)

    try:
        # NOTE: use the root logger to set the logging level
        logging.getLogger().setLevel(args.loglevel)

        log.debug("args: %s", args)
        kwargs = _get_kwargs(args)
        return args.func(**kwargs)
    except Exception as exc:  # noqa: B902
        log.critical(
            "unexpected exception caught: %r %s", type(exc).__name__, exc
        )
        log.debug("stacktrace:", exc_info=True)
        return EX_FAILURE
    except KeyboardInterrupt:
        log.warning("Keyboard interrupt received: exit the program")
        return EX_INTERRUPT



# -*- coding: utf-8 -*-
# cython: language_level=3
# distutils: language=c++

from collections import namedtuple

from .common cimport (
    GEOGRAPHICLIB_VERSION_MAJOR,
    GEOGRAPHICLIB_VERSION_MINOR,
    GEOGRAPHICLIB_VERSION_PATCH,
    GEOGRAPHICLIB_VERSION_STRING,
)

__all__ = ['lib_version_info', 'lib_version_str']


VersionInfo = namedtuple('VersionInfo', ['major', 'minor', 'micro'])


def lib_version_info() -> VersionInfo[int, int, int]:
    """Return the (major, minor, patch) version of GeographicLib.

    .. note::

       the returned information refers to the version of the
       GeographicLib library used at build time.
       The version of the shared library actually used at runtime could
       be different.
    """
    return VersionInfo(
        GEOGRAPHICLIB_VERSION_MAJOR,
        GEOGRAPHICLIB_VERSION_MINOR,
        GEOGRAPHICLIB_VERSION_PATCH,
    )


def lib_version_str() -> str:
    """Return the version string of GeographicLib.

    .. note::

       the returned information refers to the version of the
       GeographicLib library used at build time.
       The version of the shared library actually used at runtime could
       be different.
    """
    return GEOGRAPHICLIB_VERSION_STRING.decode('utf-8')



import os
import shutil
import pathlib
import datetime
import tempfile
import unittest

import numpy as np
from numpy import testing as npt

from geomodels import GravityModel, get_default_data_path
from geomodels.tests.utils import dms_to_dec


class StaticMethodsTestCase(unittest.TestCase):
    def test_default_gravity_path(self):
        self.assertIsInstance(GravityModel.default_gravity_path(), str)
        self.assertEqual(
            pathlib.Path(GravityModel.default_gravity_path()),
            pathlib.Path(get_default_data_path()).joinpath("gravity"),
        )

    def test_default_gravity_name(self):
        names = (
            "egm84",
            "egm96",
            "egm2008",
            "wgs84",
        )
        self.assertTrue(GravityModel.default_gravity_name() in names)


class InstantiationTestCase00(unittest.TestCase):
    def test_no_args(self):
        model = GravityModel()
        self.assertIsInstance(model, GravityModel)
        self.assertEqual(
            model.gravity_model_name(), GravityModel.default_gravity_name()
        )


class InstantiationTestCase(unittest.TestCase):
    MODEL_NAME = "egm96"

    def test_name(self):
        model = GravityModel(self.MODEL_NAME)
        self.assertIsInstance(model, GravityModel)
        self.assertEqual(model.gravity_model_name(), self.MODEL_NAME)

    def test_default_path(self):
        path = GravityModel.default_gravity_path()
        model = GravityModel(self.MODEL_NAME, path)
        self.assertEqual(model.gravity_model_name(), self.MODEL_NAME)
        self.assertEqual(model.gravity_model_directory(), path)

    def test_custom_path(self):
        default_path = pathlib.Path(GravityModel.default_gravity_path())
        with tempfile.TemporaryDirectory() as dirname:
            gravity_model_path = pathlib.Path(dirname) / default_path.name
            gravity_model_path.mkdir()
            for filename in default_path.glob(f"{self.MODEL_NAME}*"):
                shutil.copy(filename, gravity_model_path)

            model = GravityModel(self.MODEL_NAME, gravity_model_path)
            self.assertEqual(model.gravity_model_name(), self.MODEL_NAME)
            self.assertEqual(
                model.gravity_model_directory(), str(gravity_model_path)
            )

    def test_custom_path_from_env01(self):
        default_path = pathlib.Path(GravityModel.default_gravity_path())
        with tempfile.TemporaryDirectory() as dirname:
            gravity_model_path = pathlib.Path(dirname) / default_path.name
            gravity_model_path.mkdir()
            for filename in default_path.glob(f"{self.MODEL_NAME}*"):
                shutil.copy(filename, gravity_model_path)

            old_env = os.environ.get("GEOGRAPHICLIB_DATA")
            os.environ["GEOGRAPHICLIB_DATA"] = dirname
            try:
                model = GravityModel(self.MODEL_NAME)
                self.assertEqual(model.gravity_model_name(), self.MODEL_NAME)
                self.assertEqual(
                    pathlib.Path(model.gravity_model_directory()),
                    gravity_model_path,
                )
            finally:
                if old_env is None:
                    del os.environ["GEOGRAPHICLIB_DATA"]
                else:
                    os.environ["GEOGRAPHICLIB_DATA"] = old_env

    def test_custom_path_from_env02(self):
        default_path = pathlib.Path(GravityModel.default_gravity_path())
        with tempfile.TemporaryDirectory() as dirname:
            gravity_model_path = pathlib.Path(dirname) / default_path.name
            gravity_model_path.mkdir()
            for filename in default_path.glob(f"{self.MODEL_NAME}*"):
                shutil.copy(filename, gravity_model_path)

            old_env = os.environ.get("GEOGRAPHICLIB_GRAVITY_PATH")
            os.environ["GEOGRAPHICLIB_GRAVITY_PATH"] = str(gravity_model_path)
            try:
                model = GravityModel(self.MODEL_NAME)
                self.assertEqual(model.gravity_model_name(), self.MODEL_NAME)
                self.assertEqual(
                    model.gravity_model_directory(), str(gravity_model_path)
                )
            finally:
                if old_env is None:
                    del os.environ["GEOGRAPHICLIB_GRAVITY_PATH"]
                else:
                    os.environ["GEOGRAPHICLIB_GRAVITY_PATH"] = old_env

    def test_max_degree(self):
        max_degree = 2
        model = GravityModel(max_degree=max_degree)
        self.assertIsInstance(model, GravityModel)
        self.assertLessEqual(model.degree(), max_degree)

    def test_max_order(self):
        max_order = 2
        model = GravityModel(max_order=max_order)
        self.assertIsInstance(model, GravityModel)
        self.assertLessEqual(model.order(), max_order)


class InfoMethodsTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.name = GravityModel.default_gravity_name()
        self.datapath = GravityModel.default_gravity_path()
        self.model = GravityModel(self.name, self.datapath)

    def test_description(self):
        description = self.model.description()
        self.assertIsInstance(description, str)
        self.assertNotEqual(description, "NONE")

    def test_datetime(self):
        datestr = self.model.datetime()
        self.assertIsInstance(datestr, str)
        self.assertNotEqual(datestr, "UNKNOWN")
        date = datetime.datetime.strptime(datestr, "%Y-%m-%d")
        # date = datetime.datetime.strptime(datestr, '%Y-%m-%d %H:%M:%S')
        self.assertLess(date, datetime.datetime.now())

    def test_gravity_file(self):
        filename = self.model.gravity_file()
        self.assertIn(self.name, filename)
        self.assertIn(self.datapath, filename)

    def test_gravity_model_name(self):
        name = self.model.gravity_model_name()
        self.assertEqual(name, self.name)

    def test_gravity_model_directory(self):
        path = self.model.gravity_model_directory()
        self.assertEqual(path, self.datapath)

    def test_equatorial_radius(self):
        self.assertIsInstance(self.model.equatorial_radius(), float)
        self.assertGreater(self.model.equatorial_radius(), 5e6)

    def test_flattening(self):
        self.assertIsInstance(self.model.flattening(), float)
        self.assertGreater(self.model.flattening(), 0)
        self.assertLess(self.model.flattening(), 1)

    def test_mass_constant(self):
        self.assertIsInstance(self.model.mass_constant(), float)
        self.assertGreater(self.model.mass_constant(), 0)

    def test_reference_mass_constant(self):
        self.assertIsInstance(self.model.reference_mass_constant(), float)
        self.assertGreater(self.model.reference_mass_constant(), 0)

    def test_angular_velocity(self):
        self.assertIsInstance(self.model.angular_velocity(), float)
        self.assertGreater(self.model.angular_velocity(), 0)

    def test_degree(self):
        self.assertIsInstance(self.model.degree(), int)
        self.assertGreaterEqual(self.model.degree(), 0)
        self.assertLessEqual(self.model.degree(), 360)

    def test_order(self):
        self.assertIsInstance(self.model.order(), int)
        self.assertGreater(self.model.order(), 0)
        self.assertLessEqual(self.model.order(), 360)


class ComputationTestCase(unittest.TestCase):
    MODEL_NAME = "egm96"

    LAT = +dms_to_dec(27, 59, 17)  # 27:59:17N
    LON = +dms_to_dec(86, 55, 32)  # 86:55:32E
    HEIGHT = +8820.0  # [m]

    X = 302714.0  # [m]
    Y = 5636006.0  # [m]
    Z = 2979476.0  # [m]

    GX = -0.0002103214548  # [m/s**2]
    GY = +0.0008380880427  # [m/s**2]
    GZ = -9.7665319359240  # [m/s**2]

    MGAL = 1.0e-5  # [m/s**2]
    DELTAX = -21.03214547840420 * MGAL  # [m/s**2]
    DELTAY = +89.75849905064960 * MGAL  # [m/s**2]
    DELTAZ = -199.4330543962868 * MGAL  # [m/s**2]

    GHEIGHT = -28.743736628353

    ARCSEC = 1.0 / 3600.0
    DG01 = +208.0275697808106 * MGAL  # [m/s**2]
    XI = -18.8435137999035 * ARCSEC  # [rad]
    ETA = +4.4428027084385 * ARCSEC  # [rad]

    # $ echo 27:59:17N 86:55:32E 8820 | Gravity -n egm96 -p 13
    # -0.0002103214548 0.0008380880427 -9.7665319359240
    # $ echo 27.988055555555555 86.92555555555556 8820 | \
    # Gravity -n egm96 -p 13 -G
    # -0.0002103214548 0.0008380880427 -9.7665319359240
    # $ echo 27.988055555555555 86.92555555555556 8820 | \
    # Gravity -n egm96 -p 13 -D
    # -21.0321454784042 89.7584990506496 -199.4330543962868
    # echo 27.988055555555555 86.92555555555556 | \
    # Gravity -n egm96 -p 13 -H
    # -28.743736628353
    # $ echo 27.988055555555555 86.92555555555556 8820 | \
    # Gravity -n egm96 -p 13 -A
    # 208.0275697808106 -18.8435137999035 4.4428027084385

    def setUp(self) -> None:
        self.model = GravityModel(self.MODEL_NAME)

    def test_gravity_scalar(self):
        w, gx, gy, gz = self.model.gravity(self.LAT, self.LON, self.HEIGHT)
        npt.assert_allclose(gx, self.GX)
        npt.assert_allclose(gy, self.GY)
        npt.assert_allclose(gz, self.GZ)

    def test_disturbance_scalar(self):
        t, deltax, deltay, deltaz = self.model.disturbance(
            self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose(deltax, self.DELTAX)
        npt.assert_allclose(deltay, self.DELTAY)
        npt.assert_allclose(deltaz, self.DELTAZ)

    def test_geoid_heigt_scalar(self):
        gheight = self.model.geoid_height(self.LAT, self.LON)
        npt.assert_allclose(gheight, self.GHEIGHT)

    def test_spherical_anomaly_scalar(self):
        dg01, xi, eta = self.model.spherical_anomaly(
            self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose(dg01, self.DG01)
        npt.assert_allclose(xi, self.XI)
        npt.assert_allclose(eta, self.ETA)

    def test_w_scalar(self):
        w, gx, gy, gz = self.model.w(self.X, self.Y, self.Z)
        self.assertTrue(np.isscalar(w))
        self.assertTrue(np.isscalar(gx))
        self.assertTrue(np.isscalar(gy))
        self.assertTrue(np.isscalar(gz))

    def test_v_scalar(self):
        v, gx, gy, gz = self.model.v(self.X, self.Y, self.Z)
        self.assertTrue(np.isscalar(v))
        self.assertTrue(np.isscalar(gx))
        self.assertTrue(np.isscalar(gy))
        self.assertTrue(np.isscalar(gz))

    def test_t_components_scalar(self):
        t, gx, gy, gz = self.model.t_components(self.X, self.Y, self.Z)
        self.assertTrue(np.isscalar(t))
        self.assertTrue(np.isscalar(gx))
        self.assertTrue(np.isscalar(gy))
        self.assertTrue(np.isscalar(gz))

    def test_t_scalar(self):
        t = self.model.t(self.X, self.Y, self.Z)
        self.assertTrue(np.isscalar(t))

    def test_u_scalar(self):
        u, gammax, gammay, gammaz = self.model.u(self.X, self.Y, self.Z)
        self.assertTrue(np.isscalar(u))
        self.assertTrue(np.isscalar(gammax))
        self.assertTrue(np.isscalar(gammay))
        self.assertTrue(np.isscalar(gammaz))

    def test_phi_scalar(self):
        phi, fx, fy = self.model.phi(self.X, self.Y)
        self.assertTrue(np.isscalar(phi))
        self.assertTrue(np.isscalar(fx))
        self.assertTrue(np.isscalar(fy))


class VectorComputationTestCase(unittest.TestCase):
    MODEL_NAME = "egm96"

    LAT = np.asarray(
        [
            [+dms_to_dec(16, 46, 33), -dms_to_dec(16, 46, 43)],
            [-dms_to_dec(16, 56, 33), +dms_to_dec(16, 56, 43)],
        ]
    )
    LON = np.asarray(
        [
            [-dms_to_dec(3, 0, 34), +dms_to_dec(3, 0, 44)],
            [+dms_to_dec(3, 10, 34), -dms_to_dec(3, 10, 44)],
        ]
    )
    HEIGHT = np.asarray(
        [
            [+300, +400000],
            [+400000, +300],
        ]
    )

    X = np.asarray(
        [
            [6100258.0, 6482309.0],
            [6475723.0, 6093852.0],
        ]
    )
    Y = np.asarray(
        [
            [-320709.0, +341110.0],
            [+359341.0, -338447.0],
        ]
    )
    Z = np.asarray(
        [
            [+1829182.0, -1944859.0],
            [-1963312.0, +1847129.0],
        ]
    )

    GX = np.asarray(
        [
            [-0.0000185562383, +0.0000171252863],
            [+0.0000179414111, +0.0000255265776],
        ]
    )
    GY = np.asarray(
        [
            [-0.0000022160340, +0.0016776569530],
            [+0.0016913972644, -0.0000471425191],
        ]
    )
    GZ = np.asarray(
        [
            [-9.7839489045745, -8.6568801936947],
            [-8.6569658422907, -9.7840427535630],
        ]
    )

    MGAL = 1.0e-5  # [m/s**2]
    DELTAX = (
        np.asarray(
            [
                [-1.8556238327792, +1.7125286304307],
                [+1.7941411100352, +2.5526577571442],
            ]
        )
        * MGAL
    )
    DELTAY = (
        np.asarray(
            [
                [-0.0864975107851, -4.2925736778895],
                [-4.3994466419252, -4.5779437005530],
            ]
        )
        * MGAL
    )
    DELTAZ = (
        np.asarray(
            [
                [-24.6995051809167, -3.77044400298770],
                [-3.89353944957160, -25.6012585484130],
            ]
        )
        * MGAL
    )

    GHEIGHT = np.asarray(
        [
            [+28.707423353273, +14.886288004977],
            [+15.032497786636, +28.660865719440],
        ]
    )

    ARCSEC = 1.0 / 3600.0
    DG01 = (
        np.asarray(
            [
                [+15.7255802520963, -0.28121454847100],
                [-0.19222301397180, +16.6326172077381],
            ]
        )
        * MGAL
    )
    XI = (
        np.asarray(
            [
                [+0.0278739163708, +1.0212161827314],
                [+1.0466052916775, +0.9752134528843],
            ]
        )
        * ARCSEC
    )
    ETA = (
        np.asarray(
            [
                [+0.3912117252501, -0.4080406679571],
                [-0.4274821365431, -0.5381591725495],
            ]
        )
        * ARCSEC
    )

    def setUp(self) -> None:
        self.model = GravityModel(self.MODEL_NAME)

    def test_gravity_vector(self):
        w, gx, gy, gz = self.model.gravity(
            self.LAT.flatten(), self.LON.flatten(), self.HEIGHT.flatten()
        )
        npt.assert_allclose(gx, self.GX.flatten())
        npt.assert_allclose(gy, self.GY.flatten())
        npt.assert_allclose(gz, self.GZ.flatten())

    def test_gravity_matrix(self):
        w, gx, gy, gz = self.model.gravity(self.LAT, self.LON, self.HEIGHT)
        npt.assert_allclose(gx, self.GX)
        npt.assert_allclose(gy, self.GY)
        npt.assert_allclose(gz, self.GZ)

    def test_disturbance_vector(self):
        t, deltax, deltay, deltaz = self.model.disturbance(
            self.LAT.flatten(), self.LON.flatten(), self.HEIGHT.flatten()
        )
        npt.assert_allclose(deltax, self.DELTAX.flatten())
        npt.assert_allclose(deltay, self.DELTAY.flatten())
        npt.assert_allclose(deltaz, self.DELTAZ.flatten())

    def test_disturbance_matrix(self):
        t, deltax, deltay, deltaz = self.model.disturbance(
            self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose(deltax, self.DELTAX)
        npt.assert_allclose(deltay, self.DELTAY)
        npt.assert_allclose(deltaz, self.DELTAZ)

    def test_geoid_heigt_vector(self):
        gheight = self.model.geoid_height(
            self.LAT.flatten(), self.LON.flatten()
        )
        npt.assert_allclose(gheight, self.GHEIGHT.flatten())

    def test_geoid_heigt_matrix(self):
        gheight = self.model.geoid_height(self.LAT, self.LON)
        npt.assert_allclose(gheight, self.GHEIGHT)

    def test_spherical_anomaly_vector(self):
        dg01, xi, eta = self.model.spherical_anomaly(
            self.LAT.flatten(), self.LON.flatten(), self.HEIGHT.flatten()
        )
        npt.assert_allclose(dg01, self.DG01.flatten())
        npt.assert_allclose(xi, self.XI.flatten())
        npt.assert_allclose(eta, self.ETA.flatten())

    def test_spherical_anomaly_matrix(self):
        dg01, xi, eta = self.model.spherical_anomaly(
            self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose(dg01, self.DG01)
        npt.assert_allclose(xi, self.XI)
        npt.assert_allclose(eta, self.ETA)

    def test_w_vector(self):
        w, gx, gy, gz = self.model.w(
            self.X.flatten(), self.Y.flatten(), self.Z.flatten()
        )
        shape = self.X.flatten().shape
        self.assertEqual(w.shape, shape)
        self.assertEqual(gx.shape, shape)
        self.assertEqual(gy.shape, shape)
        self.assertEqual(gz.shape, shape)

    def test_w_matrix(self):
        w, gx, gy, gz = self.model.w(self.X, self.Y, self.Z)
        shape = self.X.shape
        self.assertEqual(w.shape, shape)
        self.assertEqual(gx.shape, shape)
        self.assertEqual(gy.shape, shape)
        self.assertEqual(gz.shape, shape)

    def test_v_vector(self):
        v, gx, gy, gz = self.model.v(
            self.X.flatten(), self.Y.flatten(), self.Z.flatten()
        )
        shape = self.X.flatten().shape
        self.assertEqual(v.shape, shape)
        self.assertEqual(gx.shape, shape)
        self.assertEqual(gy.shape, shape)
        self.assertEqual(gz.shape, shape)

    def test_v_matrix(self):
        v, gx, gy, gz = self.model.v(self.X, self.Y, self.Z)
        shape = self.X.shape
        self.assertEqual(v.shape, shape)
        self.assertEqual(gx.shape, shape)
        self.assertEqual(gy.shape, shape)
        self.assertEqual(gz.shape, shape)

    def test_t_components_vector(self):
        t, gx, gy, gz = self.model.t_components(
            self.X.flatten(), self.Y.flatten(), self.Z.flatten()
        )
        shape = self.X.flatten().shape
        self.assertEqual(t.shape, shape)
        self.assertEqual(gx.shape, shape)
        self.assertEqual(gy.shape, shape)
        self.assertEqual(gz.shape, shape)

    def test_t_components_matrix(self):
        t, gx, gy, gz = self.model.t_components(self.X, self.Y, self.Z)
        shape = self.X.shape
        self.assertEqual(t.shape, shape)
        self.assertEqual(gx.shape, shape)
        self.assertEqual(gy.shape, shape)
        self.assertEqual(gz.shape, shape)

    def test_t_vector(self):
        t = self.model.t(self.X.flatten(), self.Y.flatten(), self.Z.flatten())
        shape = self.X.flatten().shape
        self.assertEqual(t.shape, shape)

    def test_t_matrix(self):
        t = self.model.t(self.X, self.Y, self.Z)
        shape = self.X.shape
        self.assertEqual(t.shape, shape)

    def test_u_vector(self):
        u, gammax, gammay, gammaz = self.model.u(
            self.X.flatten(), self.Y.flatten(), self.Z.flatten()
        )
        shape = self.X.flatten().shape
        self.assertEqual(u.shape, shape)
        self.assertEqual(gammax.shape, shape)
        self.assertEqual(gammay.shape, shape)
        self.assertEqual(gammaz.shape, shape)

    def test_u_matrix(self):
        u, gammax, gammay, gammaz = self.model.u(self.X, self.Y, self.Z)
        shape = self.X.shape
        self.assertEqual(u.shape, shape)
        self.assertEqual(gammax.shape, shape)
        self.assertEqual(gammay.shape, shape)
        self.assertEqual(gammaz.shape, shape)

    def test_phi_vector(self):
        phi, fx, fy = self.model.phi(self.X.flatten(), self.Y.flatten())
        shape = self.X.flatten().shape
        self.assertEqual(phi.shape, shape)
        self.assertEqual(fx.shape, shape)
        self.assertEqual(fy.shape, shape)

    def test_phi_matrix(self):
        phi, fx, fy = self.model.phi(self.X, self.Y)
        shape = self.X.shape
        self.assertEqual(phi.shape, shape)
        self.assertEqual(fx.shape, shape)
        self.assertEqual(fy.shape, shape)


class ConstHeightVectorComputationTestCase(unittest.TestCase):
    MODEL_NAME = "egm96"

    LAT = np.asarray(
        [
            [+dms_to_dec(16, 46, 33), -dms_to_dec(16, 46, 43)],
            [-dms_to_dec(16, 56, 33), +dms_to_dec(16, 56, 43)],
        ]
    )
    LON = np.asarray(
        [
            [-dms_to_dec(3, 0, 34), +dms_to_dec(3, 0, 44)],
            [+dms_to_dec(3, 10, 34), -dms_to_dec(3, 10, 44)],
        ]
    )
    HEIGHT = 300.0

    GX = np.asarray(
        [
            [-0.0000185562383, +0.0000064699387],
            [+0.0000003420218, +0.0000255265776],
        ]
    )
    GY = np.asarray(
        [
            [-0.0000022160340, -0.0000728579276],
            [-0.0000753687802, -0.0000471425191],
        ]
    )
    GZ = np.asarray(
        [
            [-9.7839489045745, -9.7837475302880],
            [-9.7838283874308, -9.7840427535630],
        ]
    )

    MGAL = 1.0e-5  # [m/s**2]
    DELTAX = (
        np.asarray(
            [
                [-1.8556238327792, +0.6469938707953],
                [+0.0342021784554, +2.5526577571442],
            ]
        )
        * MGAL
    )
    DELTAY = (
        np.asarray(
            [
                [-0.0864975107851, -7.4209183975897],
                [-7.6731665518214, -4.5779437005530],
            ]
        )
        * MGAL
    )
    DELTAZ = (
        np.asarray(
            [
                [-24.6995051809167, -4.42361602478520],
                [-4.30431998082860, -25.6012585484130],
            ]
        )
        * MGAL
    )

    GHEIGHT = np.asarray(
        [
            [+28.707423353273, +14.886288004977],
            [+15.032497786636, +28.660865719440],
        ]
    )

    ARCSEC = 1.0 / 3600.0
    DG01 = (
        np.asarray(
            [
                [+15.7255802520963, -0.2935186685049],
                [-0.4570089444766, +16.6326172077381],
            ]
        )
        * MGAL
    )
    XI = (
        np.asarray(
            [
                [+0.0278739163708, +1.5627851282352],
                [+1.6159837030728, +0.9752134528843],
            ]
        )
        * ARCSEC
    )
    ETA = (
        np.asarray(
            [
                [+0.3912117252501, -0.1364024044788],
                [-0.0072106096609, -0.5381591725495],
            ]
        )
        * ARCSEC
    )

    def setUp(self) -> None:
        self.model = GravityModel(self.MODEL_NAME)

    def test_gravity_vector(self):
        w, gx, gy, gz = self.model.gravity(
            self.LAT.flatten(), self.LON.flatten(), self.HEIGHT
        )
        npt.assert_allclose(gx, self.GX.flatten())
        npt.assert_allclose(gy, self.GY.flatten())
        npt.assert_allclose(gz, self.GZ.flatten())

    def test_gravity_matrix(self):
        w, gx, gy, gz = self.model.gravity(self.LAT, self.LON, self.HEIGHT)
        npt.assert_allclose(gx, self.GX)
        npt.assert_allclose(gy, self.GY)
        npt.assert_allclose(gz, self.GZ)

    def test_disturbance_vector(self):
        t, deltax, deltay, deltaz = self.model.disturbance(
            self.LAT.flatten(), self.LON.flatten(), self.HEIGHT
        )
        npt.assert_allclose(deltax, self.DELTAX.flatten())
        npt.assert_allclose(deltay, self.DELTAY.flatten())
        npt.assert_allclose(deltaz, self.DELTAZ.flatten())

    def test_disturbance_matrix(self):
        t, deltax, deltay, deltaz = self.model.disturbance(
            self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose(deltax, self.DELTAX)
        npt.assert_allclose(deltay, self.DELTAY)
        npt.assert_allclose(deltaz, self.DELTAZ)

    def test_spherical_anomaly_vector(self):
        dg01, xi, eta = self.model.spherical_anomaly(
            self.LAT.flatten(), self.LON.flatten(), self.HEIGHT
        )
        npt.assert_allclose(dg01, self.DG01.flatten())
        npt.assert_allclose(xi, self.XI.flatten())
        npt.assert_allclose(eta, self.ETA.flatten())

    def test_spherical_anomaly_matrix(self):
        dg01, xi, eta = self.model.spherical_anomaly(
            self.LAT, self.LON, self.HEIGHT
        )
        npt.assert_allclose(dg01, self.DG01)
        npt.assert_allclose(xi, self.XI)
        npt.assert_allclose(eta, self.ETA)
